########################################################################
## SPINN DESIGN CODE
# YOUTUBE: (SPINN TV) https://www.youtube.com/spinnTv
# WEBSITE: spinncode.com
########################################################################

#                 qtpy Custom Widgets                #
#                GPL 3.0 - Kadir Aksoy                #
#   https://github.com/kadir014/pyqt5-custom-widgets  #

import random
from qtpy.QtCore import Qt, QRect, Signal, QEasingCurve, QPropertyAnimation, QSize
from qtpy.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout, QLabel, QGraphicsDropShadowEffect, QPushButton, QStyleOption, QStyle, QGraphicsOpacityEffect
from qtpy.QtGui import QColor, QPainter, QPen, QBrush, QPaintEvent

from Custom_Widgets.QCustomTheme import QCustomTheme

class LoadForm(QWidget):
    def __init__(self, form):
        super().__init__()
        # self.ui = Ui_Form()
        self.ui = form
        self.ui.setupUi(self)

        # Check the module name where ui is loaded from
        ui_module_name = form.__module__.split('.')[-1]

        # Replace "ui_" with empty string only at the start
        if ui_module_name.startswith("ui_"):
            self.ui_module_name = ui_module_name[len("ui_"):]

        self.applyThemeIcons()
        
    def applyThemeIcons(self):
        self.customTheme = QCustomTheme()
        self.customTheme.applyIcons(self, ui_file_name=self.ui_module_name)

        self.defaultTheme = self.customTheme.theme

    def paintEvent(self, e: QPaintEvent):
        super().paintEvent(e)
        
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)

class QCustomEmbededWindow(QWidget):

    closed = Signal()

    def __init__(self, parent, pos=None, title="New window", icon=None, borderRadius = 10, headerHeight = 25, animationDuration = 500, showForm = None, addWidget = None):
        super().__init__(parent)

        self.borderRadius = borderRadius
        self.setTitle(title)
        self.headerHeight = headerHeight
        self.animationDuration = animationDuration

        self.customTheme = QCustomTheme()

        if pos is None:
            pos = (random.randint(0, self.parent().width()-285), random.randint(0, self.parent().height()-190))
        self.setMinimumSize(285, 100)
        self.setGeometry(pos[0], pos[1], 285, 190)

        self.pressed = None
        self.pressed_pos = None

        self.ly = QVBoxLayout()
        self.ly.setContentsMargins(0, 0, 0, 0)
        self.ly.setAlignment(Qt.AlignTop)
        self.setLayout(self.ly)

        self.header = QWidget()
        self.header.setObjectName("header")
        self.header.setFixedHeight(self.headerHeight)
        self.header_lyt = QHBoxLayout()
        self.header_lyt.setContentsMargins(5, 5, 5, 5)
        self.header_lyt.setSpacing(5)
        self.header_lyt.setAlignment(Qt.AlignTop)
        self.header.setLayout(self.header_lyt)

        self.icon_label = QLabel()
        if icon:
            self.icon_label.setPixmap(icon.pixmap(16, 16))
            self.icon_label.setFixedSize(16, 16)
            self.header_lyt.addWidget(self.icon_label, alignment=Qt.AlignVCenter)

        self.contentwdt = QWidget()
        self.contentwdt.setObjectName("content-area")
        self.content = QVBoxLayout()
        self.contentwdt.setLayout(self.content)

        self.ly.addWidget(self.header)
        self.ly.addWidget(self.contentwdt)

        self.ttl = QLabel(self.title())

        self.close_btn = QPushButton()
        # get default icon:
        self.closeIcon = self.style().standardIcon(QStyle.SP_TitleBarCloseButton).pixmap(QSize(32, 32))
        self.close_btn.setIcon(self.closeIcon)
        self.close_btn.setObjectName("close-button")
        self.close_btn.setFixedSize(self.headerHeight, self.headerHeight)

        self.content_visible = True
        self.last_height = None

        @self.close_btn.clicked.connect
        def slot():
            self.close()
            self.closed.emit()

        self.deta_btn = QPushButton()
        self.shadeIcon = self.style().standardIcon(QStyle.SP_TitleBarShadeButton).pixmap(QSize(32, 32))
        self.unShadeIcon = self.style().standardIcon(QStyle.SP_TitleBarUnshadeButton).pixmap(QSize(32, 32))
        self.deta_btn.setIcon(self.shadeIcon)
        self.deta_btn.setObjectName("hide-button")
        self.deta_btn.setFixedSize(self.headerHeight, self.headerHeight)
        self.deta_btn.clicked.connect(self.showHideContent)
            
        self.header_lyt.addSpacing(10)
        self.header_lyt.addWidget(self.ttl, alignment=Qt.AlignVCenter)
        self.header_lyt.addWidget(self.deta_btn, alignment=Qt.AlignVCenter|Qt.AlignRight)
        self.header_lyt.addWidget(self.close_btn, alignment=Qt.AlignVCenter)

        if showForm:
            self.form = LoadForm(showForm)
            self.addWidget(self.form) 
            try:
                #older versions
                self.form.form =  self.form.ui 
                self.shownForm =  self.form.ui  
            except:
                self.shownForm = None
        
        if addWidget:
            self.addWidget(addWidget) 

    def addWidget(self, widget):
        self.content.addWidget(widget)

    def animateHeight(self, startHeight, endHeight):
        self.animation = QPropertyAnimation(self, b'maximumHeight')
        self.animation.setDuration(self.animationDuration)
        self.animation.setStartValue(startHeight)
        self.animation.setEndValue(endHeight)
        self.animation.setEasingCurve(QEasingCurve.InSine)
        self.animation.start()

    def showHideContent(self):
        if self.content_visible:
            self.content_visible = False
            self.deta_btn.setIcon(self.unShadeIcon)
            self.contentwdt.hide()
            self.animateHeight(self.height(), 0)
        else:
            self.content_visible = True
            self.contentwdt.show()
            self.deta_btn.setIcon(self.shadeIcon)
        
        self.adjustSizeToContent()

    def showEvent(self, e):
        self.adjustSizeToContent()

        """ fade in """
        opacityEffect = QGraphicsOpacityEffect(self)
        self.setGraphicsEffect(opacityEffect)
        opacityAni = QPropertyAnimation(opacityEffect, b'opacity', self)
        opacityAni.setStartValue(0)
        opacityAni.setEndValue(1)
        opacityAni.setDuration(self.animationDuration)
        opacityAni.setEasingCurve(QEasingCurve.InSine)
        opacityAni.finished.connect(lambda: self.raiseWidget(opacityAni))
        opacityAni.start()
        super().showEvent(e)

    def raiseWidget(self, opacityEffect):
        self.setShadow()
        self.raise_()

        opacityEffect.deleteLater()
    
    def adjustSizeToContent(self):
        self.adjustSize()
        # Calculate the size hint based on the content
        content_size = self.layout().sizeHint()
        # Add some padding if needed
        padding = 0
        if content_size.width() < 285:
            self.setMinimumSize(285 + padding, content_size.height() + padding)
        else:
            self.setMinimumSize(content_size.width() + padding, content_size.height() + padding)
        self.last_height = self.height()

    def setShadow(self):
        self.shadow = QGraphicsDropShadowEffect(self)
        self.shadow.setBlurRadius(8)
        self.shadow.setColor(QColor(255, 255, 255, 110) if self.customTheme.isAppDarkThemed() else QColor(0, 0, 0, 110))
        self.shadow.setOffset(0, 2)
        self.setGraphicsEffect(self.shadow)

    def __repr__(self):
        return f"<Custom.QCustomEmbededWindow(parent={self.parent()})>"

    def setTitle(self, text):
        self.titleTxt = text

    def title(self):
        return self.titleTxt
    

    def setControlsVisible(self, b):
        if not b:
            self.close_btn.hide()
            self.deta_btn.hide()

        else:
            self.close_btn.show()
            self.deta_btn.show()

    def update(self, *args, **kwargs):
        self.adjustSizeToContent()
        self.anim.update()
        super().update(*args, **kwargs)

    def mousePressEvent(self, event):
        self.startpos = self.pos()
        self.__mousePressPos = None
        self.__mouseMovePos = None
        if event.button() == Qt.LeftButton:
            self.__mousePressPos = event.globalPos()
            self.__mouseMovePos = event.globalPos()

    def mouseMoveEvent(self, event):
        if event.buttons() == Qt.LeftButton and event.y() <= self.headerHeight:
            currPos = self.mapToGlobal(self.pos())
            globalPos = event.globalPos()
            diff = globalPos - self.__mouseMovePos
            newPos = self.mapFromGlobal(currPos + diff)
            self.move(newPos)

            self.__mouseMovePos = globalPos

            self.raise_()

    def mouseReleaseEvent(self, event):
        if self.__mousePressPos is not None:
            moved = event.globalPos() - self.__mousePressPos
            if moved.manhattanLength() > 3:
                event.ignore()
                return

    def paintEvent(self, event: QPaintEvent):
        super().paintEvent(event)
        
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)
        painter.end()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'dialogcKirZF.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from qtpy.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from qtpy.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from qtpy.QtWidgets import (QApplication, QFrame, QHBoxLayout, QLabel,
    QPushButton, QScrollArea, QSizePolicy, QVBoxLayout,
    QWidget)

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName(u"Form")
        Form.resize(304, 221)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Form.sizePolicy().hasHeightForWidth())
        Form.setSizePolicy(sizePolicy)
        font = QFont()
        font.setPointSize(10)
        Form.setFont(font)
        self.verticalLayout = QVBoxLayout(Form)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(5, 5, 5, 5)
        self.widget = QWidget(Form)
        self.widget.setObjectName(u"widget")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.widget.sizePolicy().hasHeightForWidth())
        self.widget.setSizePolicy(sizePolicy1)
        self.verticalLayout_3 = QVBoxLayout(self.widget)
        self.verticalLayout_3.setSpacing(10)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.titleBar = QFrame(self.widget)
        self.titleBar.setObjectName(u"titleBar")
        self.titleBar.setFrameShape(QFrame.Shape.StyledPanel)
        self.titleBar.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout = QHBoxLayout(self.titleBar)
        self.horizontalLayout.setSpacing(10)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(10, 10, 10, 10)
        self.titleLabel = QLabel(self.titleBar)
        self.titleLabel.setObjectName(u"titleLabel")
        sizePolicy1.setHeightForWidth(self.titleLabel.sizePolicy().hasHeightForWidth())
        self.titleLabel.setSizePolicy(sizePolicy1)
        font1 = QFont()
        font1.setPointSize(10)
        font1.setBold(True)
        self.titleLabel.setFont(font1)
        self.titleLabel.setWordWrap(True)

        self.horizontalLayout.addWidget(self.titleLabel)


        self.verticalLayout_3.addWidget(self.titleBar)

        self.scrollArea = QScrollArea(self.widget)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self.mainBody = QWidget()
        self.mainBody.setObjectName(u"mainBody")
        self.mainBody.setGeometry(QRect(0, 0, 292, 103))
        self.verticalLayout_2 = QVBoxLayout(self.mainBody)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.descriptionLabel = QLabel(self.mainBody)
        self.descriptionLabel.setObjectName(u"descriptionLabel")
        sizePolicy1.setHeightForWidth(self.descriptionLabel.sizePolicy().hasHeightForWidth())
        self.descriptionLabel.setSizePolicy(sizePolicy1)
        self.descriptionLabel.setWordWrap(True)

        self.verticalLayout_2.addWidget(self.descriptionLabel)

        self.scrollArea.setWidget(self.mainBody)

        self.verticalLayout_3.addWidget(self.scrollArea)

        self.footer = QFrame(self.widget)
        self.footer.setObjectName(u"footer")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.footer.sizePolicy().hasHeightForWidth())
        self.footer.setSizePolicy(sizePolicy2)
        self.footer.setMinimumSize(QSize(200, 0))
        self.footer.setFrameShape(QFrame.Shape.StyledPanel)
        self.footer.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout_2 = QHBoxLayout(self.footer)
        self.horizontalLayout_2.setSpacing(10)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(10, 10, 10, 10)
        self.yesButton = QPushButton(self.footer)
        self.yesButton.setObjectName(u"yesButton")
        self.yesButton.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))

        self.horizontalLayout_2.addWidget(self.yesButton)

        self.cancelButton = QPushButton(self.footer)
        self.cancelButton.setObjectName(u"cancelButton")
        self.cancelButton.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))

        self.horizontalLayout_2.addWidget(self.cancelButton)


        self.verticalLayout_3.addWidget(self.footer, 0, Qt.AlignmentFlag.AlignRight)


        self.verticalLayout.addWidget(self.widget)


        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)
    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", u"Form", None))
        self.titleLabel.setText(QCoreApplication.translate("Form", u"Title", None))
        self.descriptionLabel.setText(QCoreApplication.translate("Form", u"Body", None))
        self.yesButton.setText(QCoreApplication.translate("Form", u"Accept", None))
        self.cancelButton.setText(QCoreApplication.translate("Form", u"Cancel", None))
    # retranslateUi




# coding:utf-8
# Original code: https://github.com/zhiyiYo/PyQt-Fluent-Widgets/tree/PySide6
# Edits and improvements made by Khamisi Kibet
# QT GUI BY SPINN TV(YOUTUBE)
from math import floor
import os

from qtpy.QtCore import (QEasingCurve, Qt, QPropertyAnimation, Property,
                         QParallelAnimationGroup, QSequentialAnimationGroup, QLocale)
from qtpy.QtGui import QPainter, QColor, QPalette
from qtpy.QtWidgets import QProgressBar, QStyleOption, QStyle

from Custom_Widgets.QCustomTheme import QCustomTheme


class QCustomQProgressBar(QProgressBar):
    """ Indeterminate progress bar """

    script_dir = os.path.dirname(os.path.realpath(__file__))
    WIDGET_ICON = os.path.join(script_dir, "../components/icons/linear_scale.png")
    WIDGET_TOOLTIP = "A custom animated progress bar."
    WIDGET_DOM_XML = """
    <ui language='c++'>
        <widget class='QCustomQProgressBar' name='customProgressBar'/>
    </ui>
    """
    WIDGET_MODULE = "Custom_Widgets.QCustomLoadingIndicators"

    def __init__(self, parent=None, start=True):
        super().__init__(parent=parent)

        self.customTheme = QCustomTheme()

        self._shortPos = 0
        self._longPos = 0
        self.shortBarAni = QPropertyAnimation(self, b'shortPos', self)
        self.longBarAni = QPropertyAnimation(self, b'longPos', self)

        # Customizable colors – defaults come from theme and palette:
        # For the normal state, we use the widget's highlight color.
        self._customBarColor = self.palette().color(QPalette.Highlight)
        # For error and paused, we pick different defaults based on theme.
        if self.customTheme.isAppDarkThemed():
            self._errorColor = QColor(255, 153, 164)
            self._pausedColor = QColor(252, 225, 0)
        else:
            self._errorColor = QColor(196, 43, 28)
            self._pausedColor = QColor(157, 93, 0)

        self._isError = False

        self.aniGroup = QParallelAnimationGroup(self)
        self.longBarAniGroup = QSequentialAnimationGroup(self)

        self.shortBarAni.setDuration(833)
        self.longBarAni.setDuration(1167)
        self.shortBarAni.setStartValue(0)
        self.longBarAni.setStartValue(0)
        self.shortBarAni.setEndValue(1.45)
        self.longBarAni.setEndValue(1.75)
        self.longBarAni.setEasingCurve(QEasingCurve.OutQuad)

        self.aniGroup.addAnimation(self.shortBarAni)
        self.longBarAniGroup.addPause(785)
        self.longBarAniGroup.addAnimation(self.longBarAni)
        self.aniGroup.addAnimation(self.longBarAniGroup)
        self.aniGroup.setLoopCount(-1)

        self.setFixedHeight(4)

        if start:
            self.start()

    @Property(float)
    def shortPos(self):
        return self._shortPos

    @shortPos.setter
    def shortPos(self, p):
        self._shortPos = p
        self.update()

    @Property(float)
    def longPos(self):
        return self._longPos

    @longPos.setter
    def longPos(self, p):
        self._longPos = p
        self.update()

    # Expose "paused" as a property
    @Property(bool)
    def paused(self):
        return self.aniGroup.state() == QParallelAnimationGroup.Paused

    @paused.setter
    def paused(self, isPaused: bool):
        self.aniGroup.setPaused(isPaused)
        self.update()

    # Property for the normal bar color
    @Property(QColor)
    def customBarColor(self):
        return self._customBarColor

    @customBarColor.setter
    def customBarColor(self, color: QColor):
        self._customBarColor = color
        self.update()

    # Property for the error color
    @Property(QColor)
    def errorColor(self):
        return self._errorColor

    @errorColor.setter
    def errorColor(self, color: QColor):
        self._errorColor = color
        self.update()

    # Property for the paused color
    @Property(QColor)
    def pausedColor(self):
        return self._pausedColor

    @pausedColor.setter
    def pausedColor(self, color: QColor):
        self._pausedColor = color
        self.update()

    def start(self):
        self.shortPos = 0
        self.longPos = 0
        self.aniGroup.start()
        self.update()

    def stop(self):
        self.aniGroup.stop()
        self.shortPos = 0
        self.longPos = 0
        self.update()

    def isStarted(self):
        return self.aniGroup.state() == QParallelAnimationGroup.Running

    def pause(self):
        self.aniGroup.pause()
        self.update()

    def resume(self):
        if self.aniGroup.state() == QParallelAnimationGroup.Paused:
            self.aniGroup.resume()
        
        self.update()

    def setPaused(self, isPaused: bool):
        self.aniGroup.setPaused(isPaused)
        self.update()

    def isPaused(self):
        return self.aniGroup.state() == QParallelAnimationGroup.Paused

    def error(self):
        self._isError = True
        self.aniGroup.stop()
        self.update()

    def setError(self, isError: bool):
        self._isError = isError
        if isError:
            self.error()
        else:
            self.start()

    def isError(self):
        return self._isError

    def barColor(self):
        if self.isError():
            return self._errorColor
        if self.paused:
            return self._pausedColor
        return self._customBarColor

    def paintEvent(self, e):
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        painter.setRenderHints(QPainter.Antialiasing)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)

        painter.setPen(Qt.NoPen)
        painter.setBrush(self.barColor())

        # draw short bar
        x = int((self.shortPos - 0.4) * self.width())
        w = int(0.4 * self.width())
        r = self.height() / 2
        painter.drawRoundedRect(x, 0, w, self.height(), r, r)

        # draw long bar
        x = int((self.longPos - 0.6) * self.width())
        w = int(0.6 * self.width())
        r = self.height() / 2
        painter.drawRoundedRect(x, 0, w, self.height(), r, r)



########################################################################
## IMPORTS
########################################################################
import sys
########################################################################

########################################################################
## MODULE UPDATED TO USE QTPY
########################################################################
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
########################################################################

# IMPORT QTSASS
import qtsass
########################################################################
# IMPORT OS
import os
import shutil

from Custom_Widgets.Qss.colorsystem import CreateColorVariable
########################################################################
## IMPORT WORKER
########################################################################
from Custom_Widgets.WidgetsWorker import Worker, WorkerResponse

from Custom_Widgets.Log import *

########################################################################
## COMPILE STYLESHEET CLASS
########################################################################
class CompileStyleSheet():
    def __init__(self, parent=None):
        super(CompileStyleSheet, self).__init__(parent)      

    ########################################################################
    ## APPLY COMPILED STYLESHEET
    ########################################################################
    def applyCompiledSass(self, generateIcons: bool = True, paintEntireApp: bool = True):
        settings = QSettings()
        
        qcss_folder = os.path.abspath(os.path.join(os.getcwd(), 'Qss/scss'))
        if not os.path.exists(qcss_folder):
            os.makedirs(qcss_folder)
        
        css_folder = os.path.abspath(os.path.join(os.getcwd(), 'generated-files/css/'))
        if not os.path.exists(css_folder):
            os.makedirs(css_folder)

        main_sass_path = os.path.abspath(os.path.join(os.getcwd(), 'Qss/scss/main.scss'))
        styles_sass_path = os.path.abspath(os.path.join(os.getcwd(), 'Qss/scss/_styles.scss'))
        css_path = os.path.abspath(os.path.join(os.getcwd(), 'generated-files/css/main.css'))

        variablesFile = CreateColorVariable.CreateVariables(self)

        if not os.path.exists(main_sass_path):   
            shutil.copy(os.path.abspath(os.path.join(os.path.dirname(__file__), 'scss/main.scss')), qcss_folder)  

        if not os.path.exists(styles_sass_path):   
            shutil.copy(os.path.abspath(os.path.join(os.path.dirname(__file__), 'scss/_styles.scss')), qcss_folder)  

        default_scss_path = os.path.abspath(os.path.join(os.getcwd(), 'Qss/scss/defaultStyle.scss'))
        if not os.path.exists(default_scss_path):   
            f = open(default_scss_path, 'w')
            print(f"""
            //===================================================//
            // FILE AUTO-GENERATED. PUT YOUR DEFAULT STYLES HERE. 
            // THESE STYLES WILL OVERIDE THE THEME STYLES
            //====================================================//
            
            //===================================================//
            // END //
            //====================================================//
            """, file=f)

            f.close()

        qtsass.compile_filename(main_sass_path, css_path)
        
        with open(css_path,"r") as css:
            stylesheet = css.read()

        # Create QApplication instance if it doesn't exist
        app = QApplication.instance() if QApplication.instance() else QApplication([])

        if not paintEntireApp:
            self.setStyleSheet(stylesheet)
            palette = self.palette()
            app.setStyleSheet("")
            app.setPalette(app.style().standardPalette())

        else:
            self.setStyleSheet("")
            self.setPalette(self.style().standardPalette())
            app.setStyleSheet(stylesheet)
            # newly created menus may need re-styling
            for obj in QApplication.instance().allWidgets():
                if isinstance(obj, QMenu):
                    obj.setStyleSheet(stylesheet)
        
            # palette = QPalette()
            palette = app.palette()

            # Set the background color
            try:
                # pyside2
                palette.setColor(QPalette.Background, QColor(self.theme.COLOR_BACKGROUND_1))
            except AttributeError as e:
                pass
            try:
                # pyside6
                palette.setColor(QPalette.Window, QColor(self.theme.COLOR_BACKGROUND_1))
            except AttributeError as e:
                pass
            

            # Set the text color
            palette.setColor(QPalette.Text, QColor(self.theme.COLOR_TEXT_1))

            # Set the button color
            palette.setColor(QPalette.Button, QColor(self.theme.COLOR_BACKGROUND_3))

            # Set the button text color
            palette.setColor(QPalette.ButtonText, QColor(self.theme.COLOR_TEXT_1))

            # Set the highlight color
            palette.setColor(QPalette.Highlight, QColor(self.theme.COLOR_BACKGROUND_6))

            # Set the highlight text color
            palette.setColor(QPalette.HighlightedText, QColor(self.theme.COLOR_ACCENT_1))

            # Apply the palette to the main window
            self.setPalette(palette)

            app.setPalette(palette)

        self.update()

        background_color = palette.color(QPalette.Window)
        
        # Calculate luminance using the YIQ color space formula
        luminance = (0.299 * background_color.red() + 0.587 * background_color.green() + 0.114 * background_color.blue()) / 255
        
        if luminance < 0.5:
            self.isThemeDark =  True  # Dark theme
        else:
            self.isThemeDark = False  # Light theme
            
        
        if generateIcons:
            ########################################################################
            ## GENERATE NEW ICONS
            # START WORKER
            # CURRENT THEME ICONS
            color = CreateColorVariable.getCurrentThemeInfo(self)
            normal_color = str(color["icons-color"])
            icons_folder = normal_color.replace("#", "")
            self.themeEngine.applyIcons(self, folder=icons_folder)

            self.iconsWorker = Worker(self.compileSassTheme)
            self.iconsWorker.signals.result.connect(WorkerResponse.print_output)
            self.iconsWorker.signals.finished.connect(lambda: self.themeEngine.applyIcons(self, folder=icons_folder))
            self.iconsWorker.signals.progress.connect(self.sassCompilationProgress)

            # ALL THEME ICONS
            self.allIconsWorker = Worker(self.makeAllIcons)
            self.allIconsWorker.signals.result.connect(WorkerResponse.print_output)
            self.allIconsWorker.signals.finished.connect(lambda: logInfo(self, "all icons have been checked and missing icons generated!"))
            self.allIconsWorker.signals.progress.connect(self.sassCompilationProgress)

            
            if not settings.value("ICONS-COLOR") == normal_color and color["icons-color"] is not None:     
                # Execute
                self.customWidgetsThreadpool.start(self.iconsWorker)
            else:
                self.customWidgetsThreadpool.start(self.allIconsWorker)
        

########################################################################
## ==>END
########################################################################




import os
from qtpy.QtCore import Qt, QPropertyAnimation, QSize, QEvent, Property, QEasingCurve, QRect, QPoint
from qtpy.QtWidgets import (QMdiSubWindow, QWidget, QGraphicsDropShadowEffect, QGraphicsBlurEffect,
                              QVBoxLayout, QHBoxLayout, QPushButton, QStyleOption, 
                              QStyle, QLabel, QSizePolicy, QApplication)
from qtpy.QtGui import QPainter, QColor, QPaintEvent, QShowEvent
# Import your custom utilities
from Custom_Widgets.Log import *
from Custom_Widgets.Utils import replace_url_prefix, is_in_designer, get_icon_path

# Import the animation easing curve function
from Custom_Widgets.QPropertyAnimation import returnAnimationEasingCurve

# Import the AcrylicEffect class
from Custom_Widgets.AcrylicEffect import AcrylicEffect

class QCustomHamburgerMenu(QWidget):
    """
    A customizable hamburger menu widget for Qt Designer with four-position support.
    """
    
    script_dir = os.path.dirname(os.path.realpath(__file__))
    WIDGET_ICON = os.path.join(script_dir, "components/icons/reorder.png")
    WIDGET_TOOLTIP = "A customizable hamburger menu with four-position support"
    WIDGET_DOM_XML = """
    <ui language='c++'>
        <widget class='QCustomHamburgerMenu' name='customHamburgerMenu'>
            <property name='geometry'>
                <rect>
                    <x>0</x>
                    <y>0</y>
                    <width>300</width>
                    <height>400</height>
                </rect>
            </property>
        </widget>
    </ui>
    """
    WIDGET_MODULE = "Custom_Widgets.QCustomHamburgerMenu"
    
    # String constants for position property
    POSITION_LEFT = "Left"
    POSITION_RIGHT = "Right"
    POSITION_TOP = "Top"
    POSITION_BOTTOM = "Bottom"

    def __init__(self, parent=None):
        super().__init__(parent)
        
        # Private attributes with default values
        self._position = self.POSITION_LEFT
        self._animationDuration = 250
        self._animationEasingCurve = "OutQuad"
        self._menuWidth = 300
        self._menuHeight = 400
        self._backgroundColor = QColor(255, 255, 255)
        self._shadowColor = QColor(0, 0, 0, 80)
        self._shadowBlurRadius = 30
        self._cornerRadius = 0
        self._autoHide = False
        self._overlayColor = QColor(0, 0, 0, 0)
        self._toggleButtonName = ""
        self._showButtonName = ""
        self._hideButtonName = ""
        self._toggleButtonConnected = None
        self._showButtonConnected = None
        self._hideButtonConnected = None
        self._menuHidden = True
        self._endGeometry = QRect(0, 0, self._menuWidth, self._menuHeight)
        self._sizeWrap = False
        self._center = False
        self._margin = 0 
        
        # NEW: Acrylic effect properties
        self._acrylicEnabled = False
        self._acrylicBlurRadius = 5
        self._acrylicTintColor = QColor(255, 255, 255, 0)
        self._acrylicLuminosityColor = QColor(255, 255, 255, 0)
        self._acrylicNoiseOpacity = 0.03
        
        # Initialize animation
        self.animation = QPropertyAnimation(self, b"geometry")
        self.animation.finished.connect(self._onAnimationFinished)
        
        # Create overlay
        self.overlay = QHamburgerMenuOverlay(self)
        
        # Apply initial shadow effect
        self._updateShadowEffect()
        
        # Install event filter on parent when available
        if parent:
            parent.installEventFilter(self)

        # Initialize in designer if needed
        # Window configuration - only set popup flag when NOT in designer
        if not is_in_designer(self):
            self.setWindowFlags(self.windowFlags() | Qt.FramelessWindowHint | Qt.Popup)
            self.setAttribute(Qt.WA_TranslucentBackground)
            self.hide()
            self.hideMenu(duration=100)
        else:
            # In designer, behave like a normal widget
            self.setWindowFlags(self.windowFlags() & ~Qt.Popup)
            self.setAttribute(Qt.WA_TranslucentBackground, False)

        self.blurBackground = True        
    
    def _updateShadowEffect(self):
        """Update the shadow effect without CSS."""
        current_effect = self.graphicsEffect()
        if current_effect:
            current_effect.setEnabled(False)
            self.setGraphicsEffect(None)
        
        self.shadow = QGraphicsDropShadowEffect()
        self.shadow.setBlurRadius(self._shadowBlurRadius)
        self.shadow.setColor(self._shadowColor)
        self.shadow.setOffset(0, 0)
        self.setGraphicsEffect(self.shadow)
    
    # NEW: Acrylic effect property getters and setters
    @Property(bool)
    def acrylicEnabled(self):
        return self._acrylicEnabled
    
    @acrylicEnabled.setter
    def acrylicEnabled(self, value):
        self._acrylicEnabled = bool(value)
    
    @Property(int)
    def acrylicBlurRadius(self):
        return self._acrylicBlurRadius
    
    @acrylicBlurRadius.setter
    def acrylicBlurRadius(self, value):
        self._acrylicBlurRadius = max(1, value)

    
    @Property(QColor)
    def acrylicTintColor(self):
        return self._acrylicTintColor
    
    @acrylicTintColor.setter
    def acrylicTintColor(self, value):
        self._acrylicTintColor = value
    
    @Property(QColor)
    def acrylicLuminosityColor(self):
        return self._acrylicLuminosityColor
    
    @acrylicLuminosityColor.setter
    def acrylicLuminosityColor(self, value):
        self._acrylicLuminosityColor = value
    
    @Property(float)
    def acrylicNoiseOpacity(self):
        return self._acrylicNoiseOpacity
    
    @acrylicNoiseOpacity.setter
    def acrylicNoiseOpacity(self, value):
        self._acrylicNoiseOpacity = max(0.0, min(1.0, value))
    
    def _stringToPosition(self, position_str):
        """Convert string position to normalized format."""
        position_str = str(position_str).lower().strip()
        
        if position_str in ["left", "0"]:
            return self.POSITION_LEFT
        elif position_str in ["right", "1"]:
            return self.POSITION_RIGHT
        elif position_str in ["top", "2"]:
            return self.POSITION_TOP
        elif position_str in ["bottom", "3"]:
            return self.POSITION_BOTTOM
        else:
            return self.POSITION_LEFT

    def adjustSizeToContent(self):
        """Adjust size to fit content and update positioning."""
        if not self.parent() or is_in_designer(self):
            return
            
        # Calculate the size hint based on the content
        content_size = self.sizeHint()
        
        # Automatically resize to fit content
        self.adjustSize()
        
        # Update dimensions for sizeWrap - ALWAYS use content size when sizeWrap is True
        if self._sizeWrap:
            # For ALL positions, use content size instead of parent dimensions
            self._menuWidth = content_size.width()
            self._menuHeight = content_size.height()
        
        self._updatePosition()

    def _calculateMenuGeometry(self, parent_rect, is_hidden=False):
        """Calculate menu geometry based on current settings."""
        # Calculate actual dimensions based on sizeWrap
        if self._sizeWrap:
            # When sizeWrap is True, ALWAYS use menuWidth/menuHeight (which are set to content size)
            actual_width = self._menuWidth
            actual_height = self._menuHeight
        else:
            # When sizeWrap is False, use parent dimensions for top/bottom positions
            actual_width = self._menuWidth if self._position in [self.POSITION_LEFT, self.POSITION_RIGHT] else parent_rect.width()
            actual_height = self._menuHeight if self._position in [self.POSITION_TOP, self.POSITION_BOTTOM] else parent_rect.height()

        # Calculate positioning with center option
        if self._position == self.POSITION_LEFT:
            if self._center:
                y_pos = (parent_rect.height() - actual_height) // 2
            else:
                y_pos = self._margin
            if is_hidden:
                return QRect(-actual_width, y_pos, actual_width, actual_height)
            else:
                return QRect(self._margin, y_pos, actual_width, actual_height)

        elif self._position == self.POSITION_RIGHT:
            if self._center:
                y_pos = (parent_rect.height() - actual_height) // 2
            else:
                y_pos = self._margin
            if is_hidden:
                return QRect(parent_rect.width(), y_pos, actual_width, actual_height)
            else:
                return QRect(parent_rect.width() - actual_width - self._margin, y_pos, actual_width, actual_height)

        elif self._position == self.POSITION_TOP:
            if self._center:
                x_pos = (parent_rect.width() - actual_width) // 2
            else:
                x_pos = self._margin
            if is_hidden:
                return QRect(x_pos, -actual_height, actual_width, actual_height)
            else:
                return QRect(x_pos, self._margin, actual_width, actual_height)

        elif self._position == self.POSITION_BOTTOM:
            if self._center:
                x_pos = (parent_rect.width() - actual_width) // 2
            else:
                x_pos = self._margin
            if is_hidden:
                return QRect(x_pos, parent_rect.height(), actual_width, actual_height)
            else:
                return QRect(x_pos, parent_rect.height() - actual_height - self._margin, actual_width, actual_height)

        return QRect(0, 0, actual_width, actual_height)

    # Property getters and setters
    @Property(str)
    def position(self):
        return self._position
    
    @position.setter
    def position(self, value):
        normalized_position = self._stringToPosition(value)
        
        if normalized_position in [self.POSITION_LEFT, self.POSITION_RIGHT, 
                                 self.POSITION_TOP, self.POSITION_BOTTOM]:
            self._position = normalized_position
            self._updatePosition()
        else:
            raise ValueError('Position must be "Left", "Right", "Top", or "Bottom"')
    
    @Property(int)
    def animationDuration(self):
        return self._animationDuration
    
    @animationDuration.setter
    def animationDuration(self, value):
        if value > 0:
            self._animationDuration = value
        else:
            raise ValueError("Animation duration must be positive")
    
    @Property(str)
    def animationEasingCurve(self):
        return self._animationEasingCurve
    
    @animationEasingCurve.setter
    def animationEasingCurve(self, value):
        self._animationEasingCurve = value
       
    @Property(int)
    def menuWidth(self):
        return self._menuWidth
    
    @menuWidth.setter
    def menuWidth(self, value):
        if value > 0:
            self._menuWidth = value
            # Only update position if sizeWrap is False
            if not self._sizeWrap:
                self._updatePosition()
        else:
            raise ValueError("Menu width must be positive")
    
    @Property(int)
    def menuHeight(self):
        return self._menuHeight
    
    @menuHeight.setter
    def menuHeight(self, value):
        if value > 0:
            self._menuHeight = value
            # Only update position if sizeWrap is False
            if not self._sizeWrap:
                self._updatePosition()
        else:
            raise ValueError("Menu height must be positive")
    
    @Property(bool)
    def sizeWrap(self):
        return self._sizeWrap
    
    @sizeWrap.setter
    def sizeWrap(self, value):
        self._sizeWrap = bool(value)
        if self._sizeWrap:
            # When enabling sizeWrap, adjust to content
            self.adjustSizeToContent()
        else:
            # When disabling sizeWrap, revert to configured dimensions
            self._updatePosition()
    
    @Property(bool)
    def center(self):
        return self._center
    
    @center.setter
    def center(self, value):
        self._center = bool(value)
        self._updatePosition()
    
    @Property(int)
    def margin(self):
        return self._margin
    
    @margin.setter
    def margin(self, value):
        self._margin = max(0, value)
        self._updatePosition()
    
    @Property(QColor)
    def backgroundColor(self):
        return self._backgroundColor
    
    @backgroundColor.setter
    def backgroundColor(self, value):
        self._backgroundColor = value
    
    @Property(QColor)
    def shadowColor(self):
        return self._shadowColor
    
    @shadowColor.setter
    def shadowColor(self, value):
        self._shadowColor = value
        self._updateShadowEffect()
    
    @Property(int)
    def shadowBlurRadius(self):
        return self._shadowBlurRadius
    
    @shadowBlurRadius.setter
    def shadowBlurRadius(self, value):
        self._shadowBlurRadius = value
        self._updateShadowEffect()
    
    @Property(int)
    def cornerRadius(self):
        return self._cornerRadius
    
    @cornerRadius.setter
    def cornerRadius(self, value):
        self._cornerRadius = max(0, value)
    
    @Property(bool)
    def autoHide(self):
        return self._autoHide
    
    @autoHide.setter
    def autoHide(self, value):
        self._autoHide = bool(value)
    
    @Property(QColor)
    def overlayColor(self):
        return self._overlayColor
    
    @overlayColor.setter
    def overlayColor(self, value):
        self._overlayColor = value
        if hasattr(self, 'overlay'):
            self.overlay.updateOverlayColor()
    
    @Property(str)
    def toggleButtonName(self):
        return self._toggleButtonName
    
    @toggleButtonName.setter
    def toggleButtonName(self, name):
        self._toggleButtonName = name
        toggleButton = self.getButtonByName(name)

        if self._toggleButtonConnected:
            self._toggleButtonConnected.disconnect()

        if toggleButton:
            toggleButton.clicked.connect(self.toggleMenu)
            self._toggleButtonConnected = toggleButton
        
        else:
            logWarning(f"Toggle button '{name}' not found in parent hierarchy.")
    
    @Property(str)
    def showButtonName(self):
        return self._showButtonName
    
    @showButtonName.setter
    def showButtonName(self, name):
        self._showButtonName = name
        showButton = self.getButtonByName(name)
    
        if self._showButtonConnected:
            self._showButtonConnected.disconnect()

        if showButton:
            showButton.clicked.connect(self.showMenu)
            self._showButtonConnected = showButton

        else:
            logWarning(f"Show button '{name}' not found in parent hierarchy.")
    
    @Property(str)
    def hideButtonName(self):
        return self._hideButtonName
    
    @hideButtonName.setter
    def hideButtonName(self, name):
        self._hideButtonName = name
        hideButton = self.getButtonByName(name)
    
        if self._hideButtonConnected:
            self._hideButtonConnected.disconnect()
        
        if hideButton:
            hideButton.clicked.connect(self.hideMenu)
            self._hideButtonConnected = hideButton

        
        else:
            logWarning(f"Hide button '{name}' not found in parent hierarchy.")

    def _updatePosition(self):
        """Update the menu position and size based on current settings."""
        if not self.parent() or is_in_designer(self):
            return

        # Do NOT override geometry while animating
        if self.animation.state() == QPropertyAnimation.Running:
            return

        parent_rect = self.parent().rect()

        # Calculate geometry based on current state
        if self._menuHidden:
            geometry = self._calculateMenuGeometry(parent_rect, is_hidden=True)
        else:
            geometry = self._calculateMenuGeometry(parent_rect, is_hidden=False)

        self.setGeometry(geometry)
        self._endGeometry = geometry

    
    def showMenu(self):
        """Animate the menu opening towards center."""
        if not self.parent() and not is_in_designer(self):
            return
            
        if is_in_designer(self):
            self.show()
            return
        
        self._menuHidden = False

        # Re-create overlay if deleted
        if self.overlay is None:
            self.overlay = QHamburgerMenuOverlay(self)
        
        # Adjust size to content if sizeWrap is enabled
        if self._sizeWrap:
            self.adjustSizeToContent()
        
        # Stop any ongoing animation
        if self.animation.state() == QPropertyAnimation.Running:
            self.animation.stop()
            
        parent_rect = self.parent().rect()
        
        # Calculate start and end geometries
        start_geometry = self._calculateMenuGeometry(parent_rect, is_hidden=True)
        end_geometry = self._calculateMenuGeometry(parent_rect, is_hidden=False)
        
        # Set initial position
        self.setGeometry(start_geometry)
        self.overlay.updateAcrylicEffect()
        self.show()
        
        # Show and position overlay FIRST (so it's at the bottom)
        if self.parent():
            self.overlay.setParent(self.parent())
            self.overlay.setGeometry(0, 0, parent_rect.width(), parent_rect.height())
            self.overlay.show()
            
            # CRITICAL: Ensure menu is on top of overlay
            self.raise_()
        
        # Animate opening
        self.animation.setDuration(self._animationDuration)
        self.animation.setEasingCurve(returnAnimationEasingCurve(self._animationEasingCurve))
        self.animation.setStartValue(start_geometry)
        self.animation.setEndValue(end_geometry)
        self._endGeometry = end_geometry
        self.animation.start()
    
    def hideMenu(self, duration=None):
        """Animate the menu closing away from center."""
        if not self.parent() and not is_in_designer(self):
            return
            
        if is_in_designer(self):
            return
        
        self._menuHidden = True
        
        # Stop any ongoing animation
        if self.animation.state() == QPropertyAnimation.Running:
            self.animation.stop()
            
        parent_rect = self.parent().rect()
        
        # Calculate end geometry
        start_geometry = self.geometry()
        end_geometry = self._calculateMenuGeometry(parent_rect, is_hidden=True)
        
        # Animate closing
        self.animation.setDuration(duration if duration is not None else self._animationDuration)
        self.animation.setEasingCurve(returnAnimationEasingCurve(self._animationEasingCurve))
        self.animation.setStartValue(start_geometry)
        self.animation.setEndValue(end_geometry)
        self._endGeometry = end_geometry
        self.animation.start()
    
    def _onAnimationFinished(self):
        """Handle animation completion."""
        if self.animation.state() == QPropertyAnimation.Running:
            return

        if self._menuHidden:
            self.hide()
            self.overlay.hide()

            # Delete overlay on hide
            if self.overlay is not None:
                self.overlay.hide()
                self.overlay.deleteLater()
                self.overlay = None
            
        self._updatePosition()

    def toggleMenu(self):
        """Toggle the menu visibility."""
        if self._menuHidden:
            self.showMenu()
        else:
            self.hideMenu()

    def getButtonByName(self, buttonName):
        """Recursively search for a button by objectName."""
        button_classes = [QPushButton]

        try:
            from Custom_Widgets.QCustomSidebarButton import QCustomSidebarButton
            button_classes.append(QCustomSidebarButton)
        except ImportError:
            pass

        def search_children(widget):
            if widget.objectName() == buttonName and isinstance(widget, tuple(button_classes)):
                return widget
            for child in widget.children():
                if isinstance(child, QWidget):
                    result = search_children(child)
                    if result:
                        return result
            return None

        def search_parents(widget):
            parent = widget.parent()
            while parent:
                if parent.objectName() == buttonName and isinstance(parent, tuple(button_classes)):
                    return parent
                result = search_children(parent)
                if result:
                    return result
                parent = parent.parent()
            return None

        result = search_children(self)
        if result:
            return result

        return search_parents(self)
    
    def getWidget(self, name: str) -> QWidget:
        try:
            # Validate input
            if not name or not isinstance(name, str):
                raise ValueError("Widget name must be a non-empty string")
            
            # Find the widget
            widget = self.findChild(QWidget, name)
            
            if not widget:
                logWarning(f"Widget '{name}' not found in {self.objectName() or self.__class__.__name__}")
                return None
                
            return widget
            
        except ValueError as e:
            logError(f"Invalid widget name: {e}")
            raise
        except Exception as e:
            logError(f"Unexpected error while finding widget '{name}': {str(e)}")
            return None
    
    def event(self, event):
        """Handle general events."""
        if event.type() == QEvent.ParentChange:
            self._handleParentChange()
        
        return super().event(event)

    def _handleParentChange(self):
        """Handle parent change events."""
        if self.parent() and not is_in_designer(self):
            self.parent().installEventFilter(self)
            self._updatePosition()
            if hasattr(self, 'overlay'):
                self.overlay.setParent(self.parent())

    def paintEvent(self, paintEvent: QPaintEvent) -> None:
        """Paint the widget with proper styling."""
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)

        return super().paintEvent(paintEvent)
    
    def showEvent(self, event: QShowEvent) -> None:
        """Handle show events to update position."""
        # Adjust size to content when shown if sizeWrap is enabled
        if self._sizeWrap:
            self.adjustSizeToContent()

        return super().showEvent(event)
    
    def eventFilter(self, obj, event):
        """Handle parent resize events to adjust menu and overlay size."""
        if obj == self.parent() and event.type() in [QEvent.Resize, QEvent.LayoutRequest]:
            self._updatePosition()
            if self.overlay:
                self.overlay.setGeometry(0, 0, self.parent().width(), self.parent().height())
                
        return super().eventFilter(obj, event)


class QHamburgerMenuOverlay(QWidget):
    """Overlay that covers the main content when the hamburger menu is open."""
    
    def __init__(self, parent=None):
        super().__init__(parent)
        self.hamburgerMenu = parent
        self.setWindowFlags(Qt.FramelessWindowHint)
        self.setAttribute(Qt.WA_TranslucentBackground)
        self.hide()
        
        # NEW: Acrylic effect
        self.acrylicEffect = None
        self._acrylicEnabled = False

    def setParent(self, parent):
        """Set the parent widget and install event filter."""
        super().setParent(parent)
        if parent:
            parent.installEventFilter(self)
        
    def updateOverlayColor(self):
        """Update the overlay color from parent menu."""
        if self.hamburgerMenu:
            color = self.hamburgerMenu.overlayColor
            # Use stylesheet for proper rendering
            self.setStyleSheet(f"""
                background-color: rgba({color.red()}, {color.green()}, {color.blue()}, {color.alpha()});
            """)
            self.update()
    
    def updateAcrylicEffect(self):
        """Update or create the acrylic effect based on parent menu settings."""
        if not self.hamburgerMenu or is_in_designer(self):
            return
            
        # Remove existing acrylic effect
        if self.acrylicEffect:
            self.acrylicEffect = None
        
        try:
            # Create new acrylic effect if enabled
            if self.hamburgerMenu._acrylicEnabled:
                self.acrylicEffect = AcrylicEffect(
                    widget=self,
                    blurRadius=self.hamburgerMenu._acrylicBlurRadius,
                    tintColor=self.hamburgerMenu._acrylicTintColor,
                    luminosityColor=self.hamburgerMenu._acrylicLuminosityColor,
                    noiseOpacity=self.hamburgerMenu._acrylicNoiseOpacity
                )
                
                # Apply the acrylic effect
                self.acrylicEffect.applyToWidget()
            
            self._acrylicEnabled = self.hamburgerMenu._acrylicEnabled
        except:
            pass

    def eventFilter(self, obj, event):
        """Handle parent resize events to adjust overlay size."""
        if obj == self.parent() and event.type() in [QEvent.Resize, QEvent.LayoutRequest]:
            if self.parent():
                self.setGeometry(0, 0, self.parent().width(), self.parent().height())
                # Update acrylic background when resized
                # if self.acrylicEffect and self._acrylicEnabled:
                #     self.acrylicEffect.grabFromScreen(self.parent().rect())
        return super().eventFilter(obj, event)
    
    def paintEvent(self, event):
        """Paint the overlay with proper styling."""
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        if not painter.isActive():
            return
        painter.setRenderHint(QPainter.Antialiasing)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)
        
        # Only apply regular overlay color if acrylic is disabled
        if not self._acrylicEnabled and self.hamburgerMenu:
            color = self.hamburgerMenu.overlayColor
            painter.fillRect(self.rect(), color)

        super().paintEvent(event)
    
    def mousePressEvent(self, event):
        """Close the hamburger menu when overlay is clicked."""
        if self.hamburgerMenu:
            self.hamburgerMenu.hideMenu()
    
    def showEvent(self, event):
        """Setup when overlay is shown."""
        self.updateOverlayColor()
        # self.updateAcrylicEffect()  
        
        if self.parent():
            self.setGeometry(0, 0, self.parent().width(), self.parent().height())
            
            # Grab screen for acrylic effect if enabled
            if self.acrylicEffect and self._acrylicEnabled:
                self.acrylicEffect.grabFromScreen(self.parent().rect())
        
        return super().showEvent(event)
    
    def resizeEvent(self, event):
        # self.updateAcrylicEffect()  

        return super().resizeEvent(event)
    



from Custom_Widgets.QCustomSlideMenu import QCustomSlideMenu
from Custom_Widgets.QCustomQStackedWidget import QCustomQStackedWidget as QCustomStackedWidget
from Custom_Widgets.QCustomProgressIndicator import QCustomProgressIndicator as FormProgressIndicator
from Custom_Widgets.QAppSettings import QAppSettings
from Custom_Widgets.QCustomQPushButton import QCustomQPushButton
from Custom_Widgets.__init__ import *

import warnings

warnings.warn(
    "Deprecated imports. For more information, refer to the documentation at: https://github.com/KhamisiKibet/QT-PyQt-PySide-Custom-Widgets"
)



import sys
import os
import traceback
import importlib.util
from qtpy.QtWidgets import QWidget, QStyleOption, QStyle, QLabel, QVBoxLayout, QHBoxLayout
from qtpy.QtGui import QPainter
from qtpy.QtCore import Property, Qt

from Custom_Widgets.QCustomTheme import QCustomTheme
from Custom_Widgets.Utils import get_absolute_path, is_in_designer
from Custom_Widgets.Log import *

class QCustomComponentLoader(QWidget):
    """A custom widget to load and display a UI class defined in an external file."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self._layout = None

        # Initialize UI class and setup
        self.ui = None

        self._is_designer_mode = False
        self._designer_preview = False
        self._designer_initialized = False  # Track if designer mode has been initialized

        self._file_path = None
        self._form_class = None  # Add this to track the form class
        self._form_class_name = None  # Track form class name separately

        self.themeEngine = QCustomTheme()
        self.defaultTheme = self.themeEngine.theme
        self.defaultIconsColor = self.themeEngine.iconsColor
        self.themeEngine.onThemeChanged.connect(self.applyThemeIcons)

        self._applying_icon = False
        
        # Set up designer mode immediately if in designer
        if is_in_designer(self):
            self._setup_designer_mode()
        
        self.applyThemeIcons()

    def showEvent(self, event):
        """Handle the show event to ensure designer mode is set up when widget becomes visible."""
        super().showEvent(event)
        
        # Ensure designer mode is set up when widget is shown in Qt Designer
        if is_in_designer(self) and not self._designer_initialized and not self.previewComponent:
            self._setup_designer_mode()
        
        else:
            self.applyThemeIcons()

    def applyThemeIcons(self):
        if self._applying_icon:
            return
        
        if self.ui is None:
            return
        
        self._applying_icon = True
        try:
            # Check the module name where ui is loaded from
            self.ui_module_name = self.ui.__module__.split('.')[-1]

            # Replace "ui_" with empty string only at the start
            if self.ui_module_name.startswith("ui_"):
                self.ui_module_name = self.ui_module_name[len("ui_"):]

        except Exception as e:
            self.ui_module_name = ""
            logError(f"Error determining UI module name: {e}")
            logException(traceback.format_exc())

        try:
            if self._file_path:
                file_name = os.path.basename(self._file_path).split('.')[0][len("ui_"):]
                self.themeEngine.applyIcons(self.ui, ui_file_name=file_name)

            self.themeEngine.applyIcons(self.ui, ui_file_name=self.ui_module_name)
            self.currentTheme = self.themeEngine.theme

        except Exception as e:
            logError(f"Error loading theme icons for: {self} (Module: {self.ui_module_name})")
            logError(f"Error: {e}")
            logException(traceback.format_exc())  

        finally:
            self._applying_icon = False

    def loadComponent(self, formClass=None, formClassName=None, filePath=None):
        """Load the UI class based on the provided parameters."""
        # Always show designer label when in designer mode without preview
        if is_in_designer(self) and not self.previewComponent:
            self._update_designer_label()
            return

        # If in designer mode with preview but file path is invalid, show error label
        if is_in_designer(self) and self.previewComponent:
            if filePath and not os.path.isfile(get_absolute_path(filePath)):
                self._show_error_label(f"File not found: {filePath}")
                return
            elif formClassName and not self._form_class:
                self._show_error_label(f"Class not found: {formClassName}")
                return

        # Clear any existing UI
        if self.ui is not None:
            # Remove previous UI widgets and clear layout
            for i in reversed(range(self.layout().count())):
                widget_to_remove = self.layout().itemAt(i).widget()
                if widget_to_remove is not None:
                    widget_to_remove.setParent(None)  # Remove widget from layout

        # Clear any existing layout and labels
        if self.layout() is not None:
            QWidget().setLayout(self.layout())  # Reset layout

        self.themeEngine = QCustomTheme()
        self.defaultTheme = self.themeEngine.theme
        self.defaultIconsColor = self.themeEngine.iconsColor
        
        # If formClass is provided, use it directly
        if formClass is not None:
            self._form_class = formClass
            try:
                self.ui = self._form_class()  # Instantiate the class
                self.ui.setupUi(self)
            except Exception as e:
                logError(f"Error loading form class: {e}")
                if is_in_designer(self):
                    self._show_error_label(f"Error loading class: {e}")
                return

        # If filePath is provided, handle accordingly
        elif filePath is not None:
            filePath = get_absolute_path(filePath)
            self._file_path = filePath
            
            # Check if file exists
            if not os.path.isfile(filePath):
                logError(f"File not found: {filePath}")
                if is_in_designer(self):
                    self._show_error_label(f"File not found: {os.path.basename(filePath)}")
                return
            
            if formClassName is not None:
                # Load the specific class
                self._form_class = self._import_class_from_file(filePath, formClassName)
            else:
                # Auto-detect the class name from the file path
                self._form_class = self._import_class_from_file(filePath)
            
            if self._form_class:
                try:
                    self.ui = self._form_class()  # Instantiate the class
                    self.ui.setupUi(self)
                except Exception as e:
                    logError(f"Error instantiating UI class: {e}")
                    if is_in_designer(self):
                        self._show_error_label(f"Error creating UI: {e}")
                    return
            else:
                logError("Failed to load the UI class from the specified file.")
                if is_in_designer(self):
                    self._show_error_label("No valid UI class found in file")
                return
        
        self.applyThemeIcons()
        
    def _refresh_component(self):
        self.loadComponent(formClassName=self._form_class_name, filePath=self._file_path)

    def _setup_designer_mode(self):
        """Set up the widget for Qt Designer mode."""
        if not is_in_designer(self):
            return

        self._is_designer_mode = True
        self._designer_initialized = True

        # Clear any existing layout and labels
        if self.layout() is not None:
            QWidget().setLayout(self.layout())  # Reset layout

        # Layout to hold the label (for Designer mode)
        self._layout = QVBoxLayout(self)
        self.setLayout(self._layout)  # Set the layout for the widget

        # Create a label
        self.label = QLabel(self)
        self.label.setObjectName("main_label")
        self._update_designer_label()
        
        # Add label to the layout
        self._layout.addWidget(self.label, alignment=Qt.AlignCenter)

        # Optional: Set a border to indicate that it's in designer mode
        self.setStyleSheet("QWidget { border: 1px dotted red; } #main_label { border: none; background-color: rgba(0,0,0,.6); }")

    def _ensure_designer_label(self):
        """Ensure the designer label is created and visible."""
        if not is_in_designer(self) or self.previewComponent:
            return
            
        if not hasattr(self, 'label') or self.label is None or not self._designer_initialized:
            self._setup_designer_mode()
        else:
            self._update_designer_label()

    def _show_error_label(self, error_message):
        """Show an error label when component loading fails."""
        if not is_in_designer(self):
            return
            
        # Clear any existing layout and labels
        if self.layout() is not None:
            QWidget().setLayout(self.layout())  # Reset layout

        # Layout to hold the label
        layout = QVBoxLayout(self)
        
        # Create error label
        error_label = QLabel(self)
        error_label.setObjectName("error_label")
        error_label.setText(f"<b>Component Loader - Error</b><br>"
                           f"<font color='red'>{error_message}</font><br>"
                           f"<i>Check file path and class name</i>")
        error_label.setWordWrap(True)
        error_label.setAlignment(Qt.AlignCenter)
        
        # Add label to the layout
        layout.addWidget(error_label)
        
        # Set error styling
        self.setStyleSheet("QWidget { border: 1px dotted red; background-color: #ffeeee; } #error_label { border: none; background-color: rgba(0,0,0,.6)}")

    def _update_designer_label(self):
        """Update the designer label with current configuration."""
        if not hasattr(self, 'label') or not self.label:
            return
            
        # Prepare text for label based on class name and file path
        label_text = "<b>Component Loader / Container</b>"  # Default text
        
        has_config = False
        
        if self._form_class is not None:
            class_name = self._form_class.__name__
            label_text += f"<br><b>Class:</b> {class_name}"
            has_config = True

        if self._file_path:
            file_name = os.path.basename(self._file_path)
            label_text += f"<br><b>File:</b> {file_name}"
            has_config = True
            
        if self._form_class_name and not self._form_class:
            label_text += f"<br><b>Class Name:</b> {self._form_class_name}"
            has_config = True

        if not has_config:
            label_text += "<br><i>No Class or File Loaded</i>"
        else:
            # Add preview status
            preview_status = "Enabled" if self.previewComponent else "Disabled"
            label_text += f"<br><b>Preview:</b> {preview_status}"

        # Set the label text and styling
        self.label.setText(label_text)
        self.label.setWordWrap(True)

    def _import_class_from_file(self, file_path, class_name=None):
        """Dynamically import a class from a specified Python file."""
        # Ensure the file exists
        if not os.path.isfile(file_path):
            logError(f"The specified file does not exist: {file_path}")
            return None

        try:
            # Load the module from the file
            spec = importlib.util.spec_from_file_location("module.name", file_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            # Automatically detect the class from the loaded module
            ui_classes = [cls for name, cls in module.__dict__.items() if isinstance(cls, type)]

            if class_name and class_name.strip():
                # If class_name is provided, attempt to find it
                ui_class = next((cls for cls in ui_classes if cls.__name__ == class_name), None)
                if ui_class.strip():
                    return ui_class
                else:
                    logError(f"No class named '{class_name}' found in the specified file.")
                    
            # If class_name is not provided, check for a class that follows naming conventions (e.g., starts with 'Ui_')
            ui_class = next((cls for cls in ui_classes if cls.__name__.startswith("Ui_")), None)

            if ui_class is None:
                logError("No valid UI class found in the specified file.")
                return None

            return ui_class
            
        except Exception as e:
            logError(f"Error importing from file {file_path}: {e}")
            return None

    @Property(str)
    def filePath(self):
        """Property to get or set the file path."""
        return self._file_path

    @filePath.setter
    def filePath(self, value: str):
        if self._file_path != value:
            self._file_path = value
            if is_in_designer(self):
                self._ensure_designer_label()
                # Auto-load if preview is enabled
                if self.previewComponent and value:
                    self.loadComponent(filePath=value)

    @Property(str)
    def formClassName(self):
        """Property to get or set the form class name."""
        return self._form_class_name

    @formClassName.setter
    def formClassName(self, value: str):
        if self._form_class_name != value:
            self._form_class_name = value
            if is_in_designer(self):
                self._ensure_designer_label()
                # Auto-load if preview is enabled and we have file path
                if self.previewComponent and self._file_path and value:
                    self.loadComponent(formClassName=value, filePath=self._file_path)

    @Property(bool)
    def previewComponent(self):
        """Property to get or set the preview mode."""
        return self._designer_preview

    @previewComponent.setter
    def previewComponent(self, value: bool):
        if self._designer_preview != value:
            self._designer_preview = value

            # Update the display based on the new preview state
            if is_in_designer(self):
                if value:
                    # Clear designer mode and try to load the component
                    self._is_designer_mode = False
                    self._designer_initialized = False
                    if self.layout() is not None:
                        QWidget().setLayout(self.layout())
                    # Try to load the component if we have the necessary info
                    if self._file_path:
                        if self._form_class_name:
                            self.loadComponent(formClassName=self._form_class_name, filePath=self._file_path)
                        else:
                            self.loadComponent(filePath=self._file_path)
                else:
                    # Switch back to designer mode
                    self._setup_designer_mode()
                    
                self._update_designer_label()

    def paintEvent(self, e):
        """Handle the paint event to customize the appearance of the widget."""
        super().paintEvent(e)
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)

        # Ensure designer mode is set up if needed (fallback)
        if is_in_designer(self) and not self._designer_initialized and not self.previewComponent:
            self._setup_designer_mode()
        else:
            if self.defaultIconsColor != self.themeEngine.iconsColor:
                self.defaultIconsColor = self.themeEngine.iconsColor
                self.applyThemeIcons()


#!/usr/bin/env python

###
# Author: Stefan Holstein
# inspired by: https://github.com/Werkov/PyQt4/blob/master/examples/widgets/analogclock.py
# Thanks to https://stackoverflow.com/

# Updated by 

## SPINN DESIGN CODE
# YOUTUBE: (SPINN TV) https://www.youtube.com/spinnTv
# WEBSITE: spinncode.com
# GitHub : https://github.com/KhamisiKibet


## IMPORTS
import os
import math
import json

from qtpy.QtWidgets import QWidget 
from qtpy.QtGui import QPolygon, QPolygonF, QColor, QPen, QFont, QPainter, QFontMetrics, QConicalGradient, QRadialGradient, QFontDatabase
from qtpy.QtCore import Qt, QTimer, QPoint, QPointF, QRect, QSize, QObject, Signal, Property

from Custom_Widgets.Log import *

# AnalogGaugeWidget CLASS
class AnalogGaugeWidget(QWidget):
    # Icon path for the widget
    script_dir = os.path.dirname(os.path.realpath(__file__))
    WIDGET_ICON = os.path.join(script_dir, "components/icons/speed.png")
    
    # Tooltip for the widget
    WIDGET_TOOLTIP = "An analog gauge widget"
    
    # XML string for the widget
    WIDGET_DOM_XML = """
    <ui language='c++'>
        <widget class='AnalogGaugeWidget' name='analogGaugeWidget'>
            <property name='geometry'>
                <rect>
                    <x>0</x>
                    <y>0</y>
                    <width>200</width>
                    <height>200</height>
                </rect>
            </property>
        </widget>
    </ui>
    """
    WIDGET_MODULE="Custom_Widgets.AnalogGaugeWidget"

    valueChanged = Signal(int)

    def __init__(self, parent=None, min_value: int = 0, max_value: int = 1000, 
                 needle_color: QColor = QColor(0, 0, 0, 255), 
                 needle_color_drag: QColor = QColor(255, 0, 0, 255), 
                 scale_value_color: QColor = QColor(0, 0, 0, 255), 
                 display_value_color: QColor = QColor(0, 0, 0, 255),
                 center_point_color: QColor = QColor(0, 0, 0, 255)):
        super(AnalogGaugeWidget, self).__init__(parent)

        self._needle_color = needle_color
        self._needle_color_drag = needle_color_drag
        self._scale_value_color = scale_value_color
        self._display_value_color = display_value_color
        self._center_point_color = center_point_color
        self._value_needle_count = 1
        self._min_value = min_value
        self._max_value = max_value
        self._value = self._min_value
        self._value_offset = 0
        self._value_needle_snapzone = 0.05
        self._last_value = 0
        self._gauge_color_outer_radius_factor = 1
        self._gauge_color_inner_radius_factor = 0.9
        self._center_horizontal_value = 0
        self._center_vertical_value = 0
        self._scale_angle_start_value = 135
        self._scale_angle_size = 270
        self._angle_offset = 0
        self._scala_count = 10
        self._scala_subdiv_count = 5
        self._pen = QPen(QColor(0, 0, 0))
        self._scale_polygon_colors = []
        self._big_scale_marker = Qt.black
        self._fine_scale_color = Qt.black
        self._enable_scale_text = True
        self._scale_fontname = "Orbitron"
        self._initial_scale_fontsize = 14
        self._scale_fontsize = self._initial_scale_fontsize
        self._enable_value_text = True
        self._value_fontname = "Orbitron"
        self._initial_value_fontsize = 40
        self._value_fontsize = self._initial_value_fontsize
        self._text_radius_factor = 0.5
        self._enable_bar_graph = True
        self._enable_filled_polygon = True
        self._enable_center_point = True
        self._enable_fine_scaled_marker = True
        self._enable_big_scaled_marker = True
        self._needle_scale_factor = 0.8
        self._enable_needle_polygon = True
        self._units = "℃"
        self._themeNumber = 1

        QFontDatabase.addApplicationFont(os.path.join(os.path.dirname(__file__), 'fonts/Orbitron/Orbitron-VariableFont_wght.ttf'))
        
        self.update()
        self.setGaugeTheme(0)
        self.rescale_method()

    # Properties with getters and setters
    @Property(int)
    def themeNumber(self):
        return self._themeNumber
    
    @themeNumber.setter
    def themeNumber(self, value):
        self._themeNumber = value
        self.setGaugeTheme(value)
    
    @Property(str)
    def units(self):
        return self._units

    @units.setter
    def units(self, value: str):
        self._units = value
        self.update()

    def setUnits(self, value: str):
        self.units = value

    @Property(QColor)
    def needleColor(self):
        return self._needle_color

    @needleColor.setter
    def needleColor(self, color: QColor):
        self._needle_color = color
        self.update()

    def setNeedleColor(self, color: QColor):
        self.needleColor = color

    @Property(QColor)
    def needleColorOnDrag(self):
        return self._needle_color_drag

    @needleColorOnDrag.setter
    def needleColorOnDrag(self, color: QColor):
        self._needle_color_drag = color
        self.update()

    def setNeedleColorOnDrag(self, color: QColor):
        self.needleColorOnDrag=color

    @Property(QColor)
    def scaleValueColor(self):
        return self._scale_value_color

    @scaleValueColor.setter
    def scaleValueColor(self, color: QColor):
        self._scale_value_color = color
        self.update()

    def setScaleValueColor(self, color: QColor):
        self.scaleValueColor=color

    @Property(QColor)
    def fineScaleColor(self):
        return self._fine_scale_color

    @fineScaleColor.setter
    def fineScaleColor(self, color: QColor):
        self._fine_scale_color = color
        self.update()

    # SET FINE SCALE COLOR
    def setFineScaleColor(self, color: QColor):
        self.fineScaleColor = color

    @Property(QColor)
    def displayValueColor(self):
        return self._display_value_color

    @displayValueColor.setter
    def displayValueColor(self, color: QColor):
        self._display_value_color = color
        self.update()

    def setDisplayValueColor(self, color: QColor):
        self.displayValueColor=color

    @Property(QColor)
    def centerPointColor(self):
        return self._center_point_color

    @centerPointColor.setter
    def centerPointColor(self, color: QColor):
        self._center_point_color = color
        self.update()

    def setCenterPointColor(self, color: QColor):
        self.centerPointColor=color

    @Property(int)
    def value(self):
        return self._value

    @value.setter
    def value(self, value: int):
        self._value = value

        self.update()

    def setValue(self, value: int):
        self.value=value

    @Property(int)
    def minValue(self):
        return self._min_value

    @minValue.setter
    def minValue(self, value: int):
        if self._value < value:
            self._value = value
        if value >= self._max_value:
            self._min_value = self._max_value - 1
        else:
            self._min_value = value

        self.update()

    def setMinValue(self, value: int):
        self.minValue=value

    @Property(int)
    def maxValue(self):
        return self._max_value

    @maxValue.setter
    def maxValue(self, value: int):
        if self._value > value:
            self._value = value
        if value <= self._min_value:
            self._max_value = self._min_value + 1
        else:
            self._max_value = value

        self.update()

    def setMaxValue(self, value: int):
        self.maxValue=value

    @Property(QFont)
    def scaleFontFamily(self):
        return self._scale_fontname

    @scaleFontFamily.setter
    def scaleFontFamily(self, font: QFont | str):
        self._scale_fontname = font

        self.update()

    def setScaleFontFamily(self, font: QFont | str):
        self.scaleFontFamily=font

    @Property(QFont)
    def valueFontFamily(self):
        return self._value_fontname

    @valueFontFamily.setter
    def valueFontFamily(self, font: QFont | str):
        self._value_fontname = font

        self.update()

    def setValueFontFamily(self, font: QFont | str):
        self.valueFontFamily=font

    @Property(QColor)
    def bigScaleColor(self):
        return self._big_scale_marker

    @bigScaleColor.setter
    def bigScaleColor(self, color: QColor):
        self._big_scale_marker = color

        self.update()

    def setBigScaleColor(self, color: QColor):
        self._big_scale_marker = color       

    @Property(bool)
    def enableNeedlePolygon(self):
        return self._enable_needle_polygon

    @enableNeedlePolygon.setter
    def enableNeedlePolygon(self, enable: bool):
        self._enable_needle_polygon = enable

        self.update()

    def setEnableNeedlePolygon(self, enable: bool = True):
        self.enableNeedlePolygon = enable

    @Property(bool)
    def enableScaleText(self):
        return self._enable_scale_text

    @enableScaleText.setter
    def enableScaleText(self, enable: bool = True):
        self._enable_scale_text = enable

        self.update()

    def setEnableScaleText(self, enable=True):
        self.enableScaleText = enable

    @Property(bool)
    def enableBarGraph(self):
        return self._enable_bar_graph

    @enableBarGraph.setter
    def enableBarGraph(self, enable: bool = True):
        self._enable_bar_graph = enable

        self.update()

    def setEnableBarGraph(self, enable=True):
        self.enableBarGraph = enable

    @Property(bool)
    def enableValueText(self):
        return self._enable_value_text

    @enableValueText.setter
    def enableValueText(self, enable: bool = True):
        self._enable_value_text = enable

        self.update()

    def setEnableValueText(self, enable=True):
        self.enableValueText = enable

    @Property(bool)
    def enableCenterPoint(self):
        return self._enable_center_point

    @enableCenterPoint.setter
    def enableCenterPoint(self, enable: bool = True):
        self._enable_center_point = enable

        self.update()

    def setEnableCenterPoint(self, enable=True):
        self.enableCenterPoint = enable

    @Property(bool)
    def enableScalePolygon(self):
        return self._enable_filled_polygon

    @enableScalePolygon.setter
    def enableScalePolygon(self, enable: bool = True):
        self._enable_filled_polygon = enable

        self.update()

    def setEnableScalePolygon(self, enable=True):
        self.enableScalePolygon = enable

    @Property(bool)
    def enableBigScaleGrid(self):
        return self._enable_big_scaled_marker

    @enableBigScaleGrid.setter
    def enableBigScaleGrid(self, enable: bool = True):
        self._enable_big_scaled_marker = enable

        self.update()

    def setEnableBigScaleGrid(self, enable=True):
        self.enableBigScaleGrid = enable

    @Property(bool)
    def enableFineScaleGrid(self):
        return self._enable_fine_scaled_marker

    @enableFineScaleGrid.setter
    def enableFineScaleGrid(self, enable: bool = True):
        self._enable_fine_scaled_marker = enable

        self.update()

    def setEnableFineScaleGrid(self, enable=True):
        self.enableFineScaleGrid = enable

    @Property(int)
    def scalaCount(self):
        return self._scala_count

    @scalaCount.setter
    def scalaCount(self, count: int):
        if count < 1:
            count = 1
        self._scala_count = count

        self.update()

    def setScalaCount(self, count):
        self.scalaCount = count

    @Property(float)
    def scaleStartAngle(self):
        return self._scale_angle_start_value

    @scaleStartAngle.setter
    def scaleStartAngle(self, value: float):
        self._scale_angle_start_value = value

        self.update()

    def setScaleStartAngle(self, value):
        self.scaleStartAngle = value

    @Property(float)
    def totalScaleAngleSize(self):
        return self._scale_angle_size

    @totalScaleAngleSize.setter
    def totalScaleAngleSize(self, value: float):
        self._scale_angle_size = value

        self.update()

    def setTotalScaleAngleSize(self, value: float):
        self.totalScaleAngleSize = value

    @Property(float)
    def angleOffset(self):
        return self._angle_offset

    @angleOffset.setter
    def angleOffset(self, value: float):
        self._angle_offset = value

        self.update()

    def setAngleOffset(self, offset: float):
        self.angleOffset = offset

    @Property(float)
    def gaugeColorOuterRadiusFactor(self):
        return self._gauge_color_outer_radius_factor

    @gaugeColorOuterRadiusFactor.setter
    def gaugeColorOuterRadiusFactor(self, value: float):
        self._gauge_color_outer_radius_factor = float(value) / 1000

        self.update()

    def setGaugeColorOuterRadiusFactor(self, value):
        self.gaugeColorOuterRadiusFactor = value

    @Property(float)
    def gaugeColorInnerRadiusFactor(self):
        return self._gauge_color_inner_radius_factor

    @gaugeColorInnerRadiusFactor.setter
    def gaugeColorInnerRadiusFactor(self, value: float):
        self._gauge_color_inner_radius_factor = float(value) / 1000

        self.update()

    def setGaugeColorInnerRadiusFactor(self, value):
        self.gaugeColorInnerRadiusFactor = value


    def setGaugeTheme(self, Theme=0):
        script_dir = os.path.dirname(os.path.realpath(__file__))
        theme_file = os.path.join(script_dir, "components/json/QAnalogGaugeThemes.json")
        with open(theme_file, 'r') as file:
            themes_data = json.load(file)

        if Theme < len(themes_data["themes"]):
            theme = themes_data["themes"][Theme]

            if "scale_polygon_colors" in theme:
                self.scale_polygon_colors = [
                    [scale[0], QColor(scale[1])] for scale in theme["scale_polygon_colors"]
                ]

            if "needle_center_bg" in theme:
                self.needle_center_bg = [
                    [bg[0], QColor(bg[1])] for bg in theme["needle_center_bg"]
                ]

            if "outer_circle_bg" in theme:
                self.outer_circle_bg = [
                    [bg[0], QColor(bg[1])] for bg in theme["outer_circle_bg"]
                ]

            if "bigScaleMarker" in theme:
                self._big_scale_marker = QColor(theme["bigScaleMarker"])
            if "fineScaleColor" in theme:
                self._fine_scale_color = QColor(theme["fineScaleColor"])

            if "customTheme" in theme:
                custom_theme = theme["customTheme"]
                self.setCustomGaugeTheme(custom_theme)

            self.update()
        else:
            self._themeNumber = 0
            self.setGaugeTheme(0)
    
    # SET CUSTOM GAUGE THEME
    def setCustomGaugeTheme(self, custom_theme: list):
        scale_polygon_colors = []
        needle_center_bg = []
        outer_circle_bg = []

        for x, color in enumerate(custom_theme):
            scale_polygon_colors.append([float(color[0]), QColor(color[1])])
            needle_center_bg.append([float(color[0]), QColor(color[1])])
            outer_circle_bg.append([float(color[0]), QColor(color[1])])

        self.scale_polygon_colors = scale_polygon_colors
        self.needle_center_bg = needle_center_bg
        self.outer_circle_bg = outer_circle_bg

    
    def setScalePolygonColor(self, **color_positions):
        if color_positions:
            scale_polygon_colors = []

            for position, color in color_positions.items():
                scale_polygon_colors.append([float(position), QColor(str(color))])

            self.scale_polygon_colors = scale_polygon_colors
        else:
            logInfo("Custom Gauge Theme: No colors defined")

    def setNeedleCenterColor(self, **color_positions):
        if color_positions:
            needle_center_bg = []

            for position, color in color_positions.items():
                needle_center_bg.append([float(position), QColor(str(color))])

            self.needle_center_bg = needle_center_bg
        else:
            logInfo("Custom Gauge Theme: No colors defined")

    def setOuterCircleColor(self, **color_positions):
        if color_positions:
            outer_circle_bg = []

            for position, color in color_positions.items():
                outer_circle_bg.append([float(position), QColor(str(color))])

            self.outer_circle_bg = outer_circle_bg
        else:
            logInfo("Custom Gauge Theme: No colors defined")

    # RESCALE
    def rescale_method(self):
        # SET WIDTH AND HEIGHT
        if self.width() <= self.height():
            self.widget_diameter = self.width()
        else:
            self.widget_diameter = self.height()

        
        # SET NEEDLE SIZE
        self.change_value_needle_style([QPolygon([
            QPoint(4, 30),
            QPoint(-4, 30),
            QPoint(-2, - self.widget_diameter / 2 * self._needle_scale_factor),
            QPoint(0, - self.widget_diameter / 2 * self._needle_scale_factor - 6),
            QPoint(2, - self.widget_diameter / 2 * self._needle_scale_factor)
        ])])

        # SET FONT SIZE
        self.scale_fontsize = self._initial_scale_fontsize * self.widget_diameter / 400
        self.value_fontsize = self._initial_value_fontsize * self.widget_diameter / 400


    def change_value_needle_style(self, design):
        # prepared for multiple needle instrument
        self.value_needle = []
        for i in design:
            self.value_needle.append(i)

    def center_horizontal(self, value):
        self.center_horizontal_value = value

    def center_vertical(self, value):
        self.center_vertical_value = value


    # CREATE PIE
    def create_polygon_pie(self, outer_radius, inner_raduis, start, lenght, bar_graph = True):
        polygon_pie = QPolygonF()
    
        n = 360     # angle steps size for full circle
        # changing n value will causes drawing issues
        w = 360 / n   # angle per step
        # create outer circle line from "start"-angle to "start + lenght"-angle
        x = 0
        y = 0

        # todo enable/disable bar graf here
        if not self._enable_bar_graph and bar_graph:
            # float_value = ((lenght / (self._max_value - self._min_value)) * (self._value - self._min_value))
            lenght = int(round((lenght / (self._max_value - self._min_value)) * (self._value - self._min_value)))

        for i in range(lenght+1):                                              # add the points of polygon
            t = w * i + start - self._angle_offset
            x = outer_radius * math.cos(math.radians(t))
            y = outer_radius * math.sin(math.radians(t))
            polygon_pie.append(QPointF(x, y))
        # create inner circle line from "start + lenght"-angle to "start"-angle
        for i in range(lenght+1):                                              # add the points of polygon
            t = w * (lenght - i) + start - self._angle_offset
            x = inner_raduis * math.cos(math.radians(t))
            y = inner_raduis * math.sin(math.radians(t))
            polygon_pie.append(QPointF(x, y))

        # close outer line
        polygon_pie.append(QPointF(x, y))
        return polygon_pie

    def draw_filled_polygon(self, outline_pen_with=0):
        if not self.scale_polygon_colors == None:
            painter_filled_polygon = QPainter(self)
            painter_filled_polygon.setRenderHint(QPainter.Antialiasing)
            # Koordinatenursprung in die Mitte der Flaeche legen
            painter_filled_polygon.translate(self.width() / 2, self.height() / 2)

            painter_filled_polygon.setPen(Qt.NoPen)

            self._pen.setWidth(outline_pen_with)
            if outline_pen_with > 0:
                painter_filled_polygon.setPen(self._pen)

            colored_scale_polygon = self.create_polygon_pie(
                ((self.widget_diameter / 2) - (self._pen.width() / 2)) * self._gauge_color_outer_radius_factor,
                (((self.widget_diameter / 2) - (self._pen.width() / 2)) * self._gauge_color_inner_radius_factor),
                self._scale_angle_start_value, self._scale_angle_size)

            gauge_rect = QRect(QPoint(0, 0), QSize(self.widget_diameter / 2 - 1, self.widget_diameter - 1))
            if isinstance(self.scale_polygon_colors, list):
                grad = QConicalGradient(QPointF(0, 0), - self._scale_angle_size - self._scale_angle_start_value +
                                    self._angle_offset - 1)
            else:
                grad = self.scale_polygon_colors

            # todo definition scale color as array here
            for eachcolor in self.scale_polygon_colors:
                grad.setColorAt(eachcolor[0], eachcolor[1])

            painter_filled_polygon.setBrush(grad)
            painter_filled_polygon.drawPolygon(colored_scale_polygon)

    def draw_icon_image(self):
        pass

    #######################
    # BIG SCALE MARKERS
    #######################
    def draw_big_scaled_marker(self):
        my_painter = QPainter(self)
        my_painter.setRenderHint(QPainter.Antialiasing)
        # Koordinatenursprung in die Mitte der Flaeche legen
        my_painter.translate(self.width() / 2, self.height() / 2)

        # my_painter.setPen(Qt.NoPen)
        self._pen = QPen(self._big_scale_marker)
        self._pen.setWidth(2)
        # # if outline_pen_with > 0:
        my_painter.setPen(self._pen)

        my_painter.rotate(self._scale_angle_start_value - self._angle_offset)
        steps_size = (float(self._scale_angle_size) / float(self._scala_count))
        scale_line_outer_start = self.widget_diameter/2
        scale_line_lenght = (self.widget_diameter / 2) - (self.widget_diameter / 20)
        for i in range(self._scala_count+1):
            my_painter.drawLine(scale_line_lenght, 0, scale_line_outer_start, 0)
            my_painter.rotate(steps_size)

    def create_scale_marker_values_text(self):
        painter = QPainter(self)
        # painter.setRenderHint(QPainter.HighQualityAntialiasing)
        painter.setRenderHint(QPainter.Antialiasing)

        # Koordinatenursprung in die Mitte der Flaeche legen
        painter.translate(self.width() / 2, self.height() / 2)
        # painter.save()
        if not isinstance(self._value_fontname, QFont):
            font = QFont(self._scale_fontname, self.scale_fontsize, QFont.Bold)
        else:
            font = self._scale_fontname

        fm = QFontMetrics(font)

        pen_shadow = QPen()

        pen_shadow.setBrush(self._scale_value_color)
        painter.setPen(pen_shadow)

        text_radius_factor = 0.8
        text_radius = self.widget_diameter/2 * text_radius_factor

        scale_per_div = int((self._max_value - self._min_value) / self._scala_count)

        angle_distance = (float(self._scale_angle_size) / float(self._scala_count))
        for i in range(self._scala_count + 1):
            # text = str(int((self._max_value - self._min_value) / self._scala_count * i))
            text = str(int(self._min_value + scale_per_div * i))
            w = fm.width(text) + 1
            h = fm.height()
            painter.setFont(font)
            angle = angle_distance * i + float(self._scale_angle_start_value - self._angle_offset)
            x = text_radius * math.cos(math.radians(angle))
            y = text_radius * math.sin(math.radians(angle))

            text = [x - int(w/2), y - int(h/2), int(w), int(h), Qt.AlignCenter, text]
            painter.drawText(text[0], text[1], text[2], text[3], text[4], text[5])
    
    # FINE SCALE MARKERS
    def create_fine_scaled_marker(self):
        #  Description_dict = 0
        my_painter = QPainter(self)

        my_painter.setRenderHint(QPainter.Antialiasing)
        # Koordinatenursprung in die Mitte der Flaeche legen
        my_painter.translate(self.width() / 2, self.height() / 2)

        my_painter.setPen(self._fine_scale_color)
        my_painter.rotate(self._scale_angle_start_value - self._angle_offset)
        steps_size = (float(self._scale_angle_size) / float(self._scala_count * self._scala_subdiv_count))
        scale_line_outer_start = self.widget_diameter/2
        scale_line_lenght = (self.widget_diameter / 2) - (self.widget_diameter / 40)
        for i in range((self._scala_count * self._scala_subdiv_count)+1):
            my_painter.drawLine(scale_line_lenght, 0, scale_line_outer_start, 0)
            my_painter.rotate(steps_size)

    
    # VALUE TEXT
    def create_values_text(self):
        painter = QPainter(self)
        try:
            painter.setRenderHint(QPainter.HighQualityAntialiasing)
        except AttributeError:
            try:
                painter.setRenderHint(QPainter.Antialiasing)
            except AttributeError:
                # Neither hint is available; you can handle this case as needed
                pass

        painter.translate(self.width() / 2, self.height() / 2)
        if not isinstance(self._value_fontname, QFont):
            font = QFont(self._value_fontname, self.value_fontsize, QFont.Bold)
        else:
            font = self._value_fontname
        fm = QFontMetrics(font)

        pen_shadow = QPen()

        pen_shadow.setBrush(self._display_value_color)
        painter.setPen(pen_shadow)

        text_radius = self.widget_diameter / 2 * self._text_radius_factor

        # angle_distance = (float(self._scale_angle_size) / float(self._scala_count))
        text = str(int(self._value))
        w = fm.width(text) + 1
        h = fm.height()
        painter.setFont(font)

        angle_end = float(self._scale_angle_start_value + self._scale_angle_size - 360)
        angle = (angle_end - self._scale_angle_start_value) / 2 + self._scale_angle_start_value

        x = text_radius * math.cos(math.radians(angle))
        y = text_radius * math.sin(math.radians(angle))
        text = [x - int(w/2), y - int(h/2), int(w), int(h), Qt.AlignCenter, text]
        painter.drawText(text[0], text[1], text[2], text[3], text[4], text[5])

    # UNITS TEXT
    def create_units_text(self):
        painter = QPainter(self)
        try:
            painter.setRenderHint(QPainter.HighQualityAntialiasing)
        except AttributeError:
            try:
                painter.setRenderHint(QPainter.Antialiasing)
            except AttributeError:
                # Neither hint is available; you can handle this case as needed
                pass

        painter.translate(self.width() / 2, self.height() / 2)

        if not isinstance(self._value_fontname, QFont):
            font = QFont(self._value_fontname, int(self.value_fontsize / 2.5), QFont.Bold)
        else:
            font = self._value_fontname
        fm = QFontMetrics(font)

        pen_shadow = QPen()

        pen_shadow.setBrush(self._display_value_color)
        painter.setPen(pen_shadow)

        text_radius = self.widget_diameter / 2 * self._text_radius_factor

        text = str(self._units)
        w = fm.width(text) + 1
        h = fm.height()
        painter.setFont(font)

      
        angle_end = float(self._scale_angle_start_value + self._scale_angle_size + 180)
        angle = (angle_end - self._scale_angle_start_value) / 2 + self._scale_angle_start_value

        x = text_radius * math.cos(math.radians(angle))
        y = text_radius * math.sin(math.radians(angle))
        text = [x - int(w/2), y - int(h/2), int(w), int(h), Qt.AlignCenter, text]
        painter.drawText(text[0], text[1], text[2], text[3], text[4], text[5])
    
    # CENTER POINTER
    def draw_big_needle_center_point(self, diameter=30):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.translate(self.width() / 2, self.height() / 2)
        painter.setPen(Qt.NoPen)

        # create_polygon_pie(self, outer_radius, inner_raduis, start, lenght)
        colored_scale_polygon = self.create_polygon_pie(
                ((self.widget_diameter / 8) - (self._pen.width() / 2)),
                0,
                self._scale_angle_start_value, 360, False)

        grad = QConicalGradient(QPointF(0, 0), 0)

        for eachcolor in self.needle_center_bg:
            grad.setColorAt(eachcolor[0], eachcolor[1])

        painter.setBrush(grad)
        painter.drawPolygon(colored_scale_polygon)
    
    # CREATE OUTER COVER
    def draw_outer_circle(self, diameter=30):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)

        painter.translate(self.width() / 2, self.height() / 2)
        painter.setPen(Qt.NoPen)
        colored_scale_polygon = self.create_polygon_pie(
                ((self.widget_diameter / 2) - (self._pen.width())),
                (self.widget_diameter / 6),
                self._scale_angle_start_value / 10, 360, False)

        radialGradient = QRadialGradient(QPointF(0, 0), self.width())

        for eachcolor in self.outer_circle_bg:
            radialGradient.setColorAt(eachcolor[0], eachcolor[1])

        painter.setBrush(radialGradient)
        painter.drawPolygon(colored_scale_polygon)

    # NEEDLE POINTER
    def draw_needle(self):
        painter = QPainter(self)
        # painter.setRenderHint(QtGui.QPainter.HighQualityAntialiasing)
        painter.setRenderHint(QPainter.Antialiasing)
        # Koordinatenursprung in die Mitte der Flaeche legen
        painter.translate(self.width() / 2, self.height() / 2)
        painter.setPen(Qt.NoPen)
        painter.setBrush(self._needle_color)
        painter.rotate(((self._value - self._value_offset - self._min_value) * self._scale_angle_size /
                        (self._max_value - self._min_value)) + 90 + self._scale_angle_start_value)

        painter.drawConvexPolygon(self.value_needle[0])

    # ON WINDOW RESIZE
    def resizeEvent(self, event):
        self.rescale_method()

    
    # ON PAINT EVENT
    def paintEvent(self, event):
        self.draw_outer_circle()
        self.draw_icon_image()
        # colored pie area
        if self._enable_filled_polygon:
            self.draw_filled_polygon()

        # draw scale marker lines
        if self._enable_fine_scaled_marker:
            self.create_fine_scaled_marker()
        if self._enable_big_scaled_marker:
            self.draw_big_scaled_marker()

        # draw scale marker value text
        if self._enable_scale_text:
            self.create_scale_marker_values_text()

        # Display Value
        if self._enable_value_text:
            self.create_values_text()
            self.create_units_text()

        # draw needle 1
        if self._enable_needle_polygon:
            self.draw_needle()

        # Draw Center Point
        if self._enable_center_point:
            self.draw_big_needle_center_point(diameter=(self.widget_diameter / 6))

    # MOUSE EVENTS
    def setMouseTracking(self, flag):
        def recursive_set(parent):
            for child in parent.findChildren(QObject):
                try:
                    child.setMouseTracking(flag)
                except:
                    pass
                recursive_set(child)

        QWidget.setMouseTracking(self, flag)
        recursive_set(self)

    def mouseReleaseEvent(self, QMouseEvent):
        self._needle_color = self._needle_color
        self.update()

    
    ## MOUSE LEAVE EVENT
    def leaveEvent(self, event):
        self._needle_color = self._needle_color
        self.update() 

    def mouseMoveEvent(self, event):
        x, y = event.x() - (self.width() / 2), event.y() - (self.height() / 2)
        if not x == 0: 
            angle = math.atan2(y, x) / math.pi * 180
            value = (float(math.fmod(angle - self._scale_angle_start_value + 720, 360)) / \
                     (float(self._scale_angle_size) / float(self._max_value - self._min_value))) + self._min_value
            temp = value
            fmod = float(math.fmod(angle - self._scale_angle_start_value + 720, 360))
            state = 0
            if (self._value - (self._max_value - self._min_value) * self._value_needle_snapzone) <= \
                    value <= \
                    (self._value + (self._max_value - self._min_value) * self._value_needle_snapzone):
                self._needle_color = self._needle_color_drag
                # todo: evtl ueberpruefen
                #
                state = 9
                if value >= self._max_value and self.last_value < (self._max_value - self._min_value) / 2:
                    state = 1
                    value = self._max_value
                    self.last_value = self._min_value
                    self.valueChanged.emit(int(value))

                elif value >= self._max_value >= self.last_value:
                    state = 2
                    value = self._max_value
                    self.last_value = self._max_value
                    self.valueChanged.emit(int(value))


                else:
                    state = 3
                    self.last_value = value
                    self.valueChanged.emit(int(value))

                self.setValue(value)   

        self.update()         




from qtpy import QtGui, QtCore, QtWidgets
import typing

# https://github.com/baoboa/pyqt5/blob/master/examples/layouts/flowlayout.py
class QCustomFlowLayout(QtWidgets.QLayout):
    def __init__(self, parent=None, margin=0, spacing=-1):
        super().__init__(parent)

        if parent is not None:
            self.setContentsMargins(margin, margin, margin, margin)

        self.setSpacing(spacing)

        self._items = []
        self.__pending_positions = {}

    def __del__(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addItem(self, a0: QtWidgets.QLayoutItem) -> None:
        try:
            position = self.__pending_positions[a0.widget()]
            self._items.insert(position, a0)
            del self.__pending_positions[a0.widget()]
        except KeyError:
            self._items.append(a0)

    def addWidget(self, w: QtWidgets.QWidget, position: int = None, align: QtCore.Qt.AlignmentFlag = None) -> None:
        if position is not None:
            self.__pending_positions[w] = position
        if align is not None:
            frame_layout = w.layout()
            if frame_layout is not None:
                frame_layout.setAlignment(align)
        super().addWidget(w)

    def count(self):
        return len(self._items)

    def expandingDirections(self):
        return QtCore.Qt.Orientations(QtCore.Qt.Orientation(0))

    def itemAt(self, index: int) -> QtWidgets.QLayoutItem:
        if 0 <= index < len(self._items):
            return self._items[index]

        return None

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        height = self._doLayout(QtCore.QRect(0, 0, width, 0), True)
        return height

    def minimumSize(self):
        size = QtCore.QSize()

        for item in self._items:
            size = size.expandedTo(item.minimumSize())

        margin, _, _, _ = self.getContentsMargins()

        size += QtCore.QSize(2 * margin, 2 * margin)
        return size

    def removeItem(self, a0: QtWidgets.QLayoutItem) -> None:
        a0.widget().deleteLater()

    def removeWidget(self, w: QtWidgets.QWidget) -> None:
        w.deleteLater()

    def setGeometry(self, rect):
        super().setGeometry(rect)
        self._doLayout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def takeAt(self, index: int) -> QtWidgets.QLayoutItem:
        if 0 <= index < len(self._items):
            return self._items.pop(index)

        return None

    def _doLayout(self, rect, testOnly):
        """This does the layout. Dont ask me how. Source: https://github.com/baoboa/pyqt5/blob/master/examples/layouts/flowlayout.py"""
        x = rect.x()
        y = rect.y()
        line_height = 0

        for item in self._items:
            wid = item.widget()
            space_x = self.spacing() + wid.style().layoutSpacing(
                QtWidgets.QSizePolicy.PushButton,
                QtWidgets.QSizePolicy.PushButton,
                QtCore.Qt.Horizontal)
            space_y = self.spacing() + wid.style().layoutSpacing(
                QtWidgets.QSizePolicy.PushButton,
                QtWidgets.QSizePolicy.PushButton,
                QtCore.Qt.Vertical)
            next_x = x + item.sizeHint().width() + space_x
            if next_x - space_x > rect.right() and line_height > 0:
                x = rect.x()
                y = y + line_height + space_y
                next_x = x + item.sizeHint().width() + space_x
                line_height = 0

            if not testOnly:
                item.setGeometry(QtCore.QRect(QtCore.QPoint(x, y), item.sizeHint()))

            x = next_x
            line_height = max(line_height, item.sizeHint().height())

        return y + line_height - rect.y()
    
    def resizeEvent(self, event):
        super().resizeEvent(event)
        self._doLayout(self.geometry(), False)


import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name='QT-PyQt-PySide-Custom-Widgets',
    packages=[
        'Custom_Widgets',
        'Custom_Widgets.iconify',
        'Custom_Widgets.Qss',
    ],
    include_package_data=True,
    version='2.0.6',
    license="GNU General Public License v3.0",
    description='A comprehensive library of custom Qt widgets, animations, and UI components for PySide and PyQt desktop applications.',
    long_description=README,
    long_description_content_type="text/markdown",
    author='Khamisi Kibet',
    author_email='kibetkhamisi@gmail.com',
    url='https://github.com/SpinnCompany/QT-PyQt-PySide-Custom-Widgets',
    download_url='https://github.com/SpinnCompany/QT-PyQt-PySide-Custom-Widgets/archive/refs/heads/main.zip',
    keywords=[
        'PySide', 'PyQt', 'Qt', 'Custom Widgets', 'Animations', 'Desktop GUI', 'QML',
        'Designer', 'Qt Creator', 'Python GUI', 'C++', 'Modern UI', 'Component Library'
    ],
    install_requires=[
        "qtpy",
        "cairosvg",
        "qtsass",
        "matplotlib",
        "mock",
        "termcolor",
        "watchdog",
        "lxml",
        "setuptools",
        "kids-cache",
        "perlin_noise",
        "colorthief",
        "scipy",
        "Pillow",
        "rich"
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    entry_points={
        'console_scripts': [
            'Custom_Widgets=Custom_Widgets.CMD:run_command',
        ],
    },
)



from qtpy.QtWidgets import QWidget, QSizePolicy, QApplication
from qtpy.QtCore import QSize, Qt, Property
from qtpy.QtGui import QPainter, QColor, QPalette

import os

class QCustomVerticalSeparator(QWidget):
    # Meta-information for integration with Qt Designer or other uses
    script_dir = os.path.dirname(os.path.realpath(__file__))
    WIDGET_ICON = os.path.join(script_dir, "components/icons/vertical_split.png")
    WIDGET_TOOLTIP = "A custom vertical separator widget"
    WIDGET_DOM_XML = """
    <ui language='c++'>
        <widget class='QCustomVerticalSeparator' name='customVerticalSeparator'>
        </widget>
    </ui>
    """
    WIDGET_MODULE = "Custom_Widgets.QCustomVerticalSeparator"

    def __init__(self, parent=None, color=None, width=1, margin=8):
        super().__init__(parent)
        
        # If no color is passed, use the default text color from the palette
        if color is not None:
            self.setColor(color)
        
        self._width = width
        self._margin = margin
        
        # Set the size policy correctly using QSizePolicy enums (Fixed for width, Expanding for height)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Expanding)
        self.setSizePolicy(sizePolicy)

    def getColor(self):
        # Get the palette's text color (matches application theme)
        self._color = self.palette().color(QPalette.Text)
        return self._color

    def setColor(self, color):
        """Sets the color of the separator."""
        self._color = QColor(color)
        self.update()

    @Property(QColor)
    def color(self):
        return self.getColor()

    @color.setter
    def color(self, color):
        self.setColor(color)

    # Width Property (equivalent to height in horizontal separator)
    def getWidth(self):
        return self._width

    def setWidth(self, width):
        """Sets the width (thickness) of the separator."""
        self._width = width
        self.updateGeometry()

    @Property(int)
    def width(self):
        return self.getWidth()

    @width.setter
    def width(self, width):
        self.setWidth(width)

    # Margin Property
    def getMargin(self):
        return self._margin

    def setMargin(self, margin):
        """Sets the horizontal margin around the separator line."""
        self._margin = margin
        self.updateGeometry()

    @Property(int)
    def margin(self):
        return self.getMargin()

    @margin.setter
    def margin(self, margin):
        self.setMargin(margin)

    def sizeHint(self):
        """Returns the suggested size for the separator."""
        return QSize(self._width + 2 * self._margin, 100)

    def minimumSizeHint(self):
        """Returns the minimum size hint for the separator."""
        return QSize(self._width + 2 * self._margin, 20)

    def paintEvent(self, event):
        """Overrides the paint event to draw the separator."""
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        
        # Set the pen and brush color for the line
        self._color = self.getColor()
        painter.setPen(Qt.NoPen)
        painter.setBrush(self._color)
        
        # Calculate the position and dimensions for the separator
        x = self._margin
        rect = self.rect()
        painter.drawRect(x, 0, self._width, rect.height())

        painter.end()


# -*- coding: utf-8 -*-

import os

# FIX: DLL import issues
def set_dll_search_path():
    # Python 3.8 no longer searches for DLLs in PATH, so we have to add
    # everything in PATH manually. Note that unlike PATH add_dll_directory
    # has no defined order, so if there are two cairo DLLs in PATH we
    # might get a random one.
    if os.name != "nt" or not hasattr(os, "add_dll_directory"):
        return
    for p in os.environ.get("PATH", "").split(os.pathsep):
        try:
            os.add_dll_directory(p)
        except OSError:
            pass


set_dll_search_path()

import cairosvg
import codecs
import subprocess
import shutil
from urllib.parse import urlparse
import __main__

from Custom_Widgets.Qss.colorsystem import *
import qtpy
from qtpy.QtCore import *

from Custom_Widgets.Log import *

from PySide6.QtGui import QColor

class NewIconsGenerator(QObject):
    def __init__(self, arg):
        super(NewIconsGenerator, self).__init__()
        self.arg = arg

    def getAllFolders(base_folder):
        all_folders = []
        for root, dirs, files in os.walk(base_folder):
            for dir_name in dirs:
                folder_path = os.path.relpath(os.path.join(root, dir_name), base_folder)
                all_folders.append(folder_path)
        return all_folders

    def getAllSvgFiles(base_folder):
        svg_files = []
        for file_name in os.listdir(base_folder):
            if file_name.lower().endswith('.svg'):
                file_path = os.path.join(base_folder, file_name)
                svg_files.append(file_path)
        return svg_files
    
    def generateIcons(progress_callback, iconsColor, suffix, iconsFolder="", createQrc=False, output_width=None, output_height=None):
        # Base folder
        base_folder = os.path.dirname(__file__)
        icons_folder_base = os.path.join(base_folder, 'icons')

        svg_color = "#ffffff"

        # Get a list of all folders inside 'icons'
        folders = NewIconsGenerator.getAllFolders(icons_folder_base)
        new_icon_made = False
        qrc_content = f'<RCC>\n'

        qrc_folder_path = os.path.abspath(os.path.join(os.getcwd(), f'Qss/icons'))
        
        for folder in folders:
            qrc_prefix = (folder+suffix).replace("/", "_")
            qrc_prefix = (folder+suffix).replace("\\", "_")
            qrc_content += f'  <qresource prefix="{qrc_prefix}">\n'
            
            icons_folder_path = os.path.abspath(os.path.join(os.getcwd(), f'Qss/icons/{iconsFolder}/{folder}'))

            if not os.path.exists(icons_folder_path):
                os.makedirs(icons_folder_path)

            folder_path = os.path.join(icons_folder_base, folder)
            list_of_files = NewIconsGenerator.getAllSvgFiles(folder_path)
            total_icons = len(list_of_files)

            for index, file_path in enumerate(list_of_files):
                file_name = os.path.basename(urlparse(file_path).path).replace(".svg", f"{suffix}.png")
                output_path = os.path.abspath(os.path.join(icons_folder_path, file_name))

                qrc_content += f'    <file>icons/{folder}/{file_name}</file>\n'
                progress_callback.emit(int((index / total_icons) * 100))

                if os.path.exists(output_path):
                    continue

                try:
                    with codecs.open(file_path, encoding='utf-8', errors='ignore') as f:
                        content = f.read()

                        new_svg = content.replace(svg_color, iconsColor)
                        new_bytes = str.encode(new_svg)

                        if output_height is not None and output_width is not None:
                            cairosvg.svg2png(bytestring=new_bytes, write_to=output_path, output_width=output_width, output_height=output_height)
                        
                        else:
                            cairosvg.svg2png(bytestring=new_bytes, write_to=output_path)

                        new_icon_made = True

                except Exception as e:
                    print(f"Error processing {file_path}: {e}")

            qrc_content += f'  </qresource>\n'

        qrc_content += f'</RCC>\n'
        qrc_file_path = os.path.abspath(os.path.join(qrc_folder_path, f'{suffix}_icons.qrc'))

        if (createQrc and new_icon_made) or not os.path.exists(qrc_file_path):
            NewIconsGenerator.createQrcFile(qrc_content, qrc_file_path)
            # Convert qrc to py 
            # qrc_output_path = qrc_file_path.replace(".qrc", "_rc.py")
            # qrc_output_path = qrc_output_path.replace("Qss/", "") #linux
            # qrc_output_path = qrc_output_path.replace("Qss\\", "") #windows
            # NewIconsGenerator.qrcToPy(qrc_file_path, qrc_output_path)

    def generateNewIcons(self, progress_callback): 
        # Icons color
        color = CreateColorVariable.getCurrentThemeInfo(self)
        normal_color = str(color["icons-color"])
        
        settings = QSettings()
        if not settings.value("ICONS-COLOR") == normal_color and color["icons-color"] is not None:
            logInfo(self, ("Current icons color ", settings.value("ICONS-COLOR"), "New icons color", normal_color))
            logInfo(self, "Generating icons for your theme, please wait. This may take long")
            
            settings.setValue("ICONS-COLOR", normal_color)
            iconsFolderName = normal_color.replace("#", "")
            self.themeEngine.applyIcons(self, folder=iconsFolderName)

            if not self.checkForMissingicons:
                # emit 100% progress
                progress_callback.emit(100)
                return 
            # Create normal icons
            NewIconsGenerator.generateIcons(progress_callback, normal_color, "", iconsFolderName, createQrc = False)

            logInfo(self, ("DONE: Current icons color ", settings.value("ICONS-COLOR")))
                

    def generateAllIcons(self, progress_callback):
        if not self.checkForMissingicons:
            # emit 100% progress
            progress_callback.emit(100)
            return
        
        themes = self.ui.themes
        def get_theme_color(theme):
            if hasattr(theme, "iconsColor") and theme.iconsColor != "":
                return theme.iconsColor
            if hasattr(theme, "accentColor") and theme.accentColor != "":
                return theme.accentColor
            return ""

        for theme in themes:
            color = get_theme_color(theme)
            try:
                if color == "" or color == settings.value("ICONS-COLOR"):
                    continue
            except:
                pass

            logInfo(self, f"Checking icons for {theme.name} theme. Icons color: {color}")

            iconsFolderName = color.replace("#", "")
            NewIconsGenerator.generateIcons(progress_callback, color, "", iconsFolderName)
        
        # then make icons for qt designer
        logInfo(self, f"Checking icons for qt designer app.")
        if settings.value("DESIGNER-ICONS-COLOR") is not None:
            NewIconsGenerator.generateIcons(progress_callback, settings.value("DESIGNER-ICONS-COLOR"), "", "icons", createQrc=True, output_width=24, output_height=24)
        else:
            NewIconsGenerator.generateIcons(progress_callback, "#000", "", "icons", createQrc=True, output_width=24, output_height=24)

    def createQrcFile(contents, filePath):
        # Save QRC content to a file
        with open(filePath, 'w', encoding='utf-8') as qrc_file:
            qrc_file.write(contents)

        # print(f'QRC file generated: {filePath}')
    
    def qrcToPy(qrcFile, pyFile):
        """
        Convert a Qt Resource Collection (qrc) file to a Python file.

        Parameters:
        - qrc_file (str): Path to the input qrc file.
        - py_file (str): Path to the output py file.
        """
        try:
            if qtpy.API_NAME == "PyQt5":
                rcc_command = 'pyrcc5'
            elif qtpy.API_NAME == "PyQt6":
                rcc_command = 'pyrcc6'
            elif qtpy.API_NAME == "PySide2":
                rcc_command = 'pyside2-rcc'
            elif qtpy.API_NAME == "PySide6":
                rcc_command = 'pyside6-rcc'
            else:
                raise Exception("Error: Unknown QT binding/API Name", qtpy.API_NAME)

            print(f'{rcc_command} "{qrcFile}" -o "{pyFile}"')
            subprocess.run(f'{rcc_command} "{qrcFile}" -o "{pyFile}"')
            
        except Exception as e:
            print("Error converting qrc to py:", e)

    def uiToPy(uiFile, pyFile):
        """
        Convert a Qt UI file to a Python file.

        Parameters:
        - uiFile (str): Path to the input UI file.
        - pyFile (str): Path to the output Python file.
        """
        try:
            if qtpy.API_NAME == "PyQt5":
                pyuic_command = 'pyuic5'
            elif qtpy.API_NAME == "PyQt6":
                pyuic_command = 'pyuic6'
            elif qtpy.API_NAME == "PySide2":
                pyuic_command = 'pyside2-uic'
            elif qtpy.API_NAME == "PySide6":
                pyuic_command = 'pyside6-uic'
            else:
                raise Exception("Error: Unknown Qt binding/API Name", qtpy.API_NAME)

            os.system(f'{pyuic_command} "{uiFile}" -o "{pyFile}"')

        except Exception as e:
            print("Error converting ui to py:", e)

    def renameFolder(old_name, new_name):
        try:
            # Check if the destination directory exists
            if os.path.exists(new_name):
                # Remove the destination directory if it exists
                shutil.rmtree(new_name)

            # Rename the folder
            os.rename(old_name, new_name)
        except Exception as e:
            pass









## SPINN DESIGN CODE
# YOUTUBE: (SPINN TV) https://www.youtube.com/spinnTv
# WEBSITE: spinncode.com

## MODULE UPDATED TO USE QT.PY
from qtpy.QtCore import Qt, QEasingCurve, QPoint, Slot, QParallelAnimationGroup, QPropertyAnimation, QAbstractAnimation, QTimeLine, Property
from qtpy.QtGui import QPainter, QPixmap
from qtpy.QtWidgets import QStackedWidget, QWidget, QGraphicsOpacityEffect, QStyleOption, QStyle, QPushButton
import os

from Custom_Widgets.QPropertyAnimation import returnAnimationEasingCurve, returnQtDirection

class QCustomQStackedWidget(QStackedWidget):
    # Define the XML metadata and icon for Qt Designer
    script_dir = os.path.dirname(os.path.realpath(__file__))
    WIDGET_ICON = os.path.join(script_dir, "components/icons/layers.png")
    WIDGET_TOOLTIP = "A custom QStackedWidget with transitions"
    WIDGET_DOM_XML = """
    <ui language='c++'>
        <widget class='QCustomQStackedWidget' name='customQStackedWidget'>
        </widget>
    </ui>
    """
    WIDGET_MODULE = "Custom_Widgets.QCustomQStackedWidget"

    def __init__(self, parent=None):
        super(QCustomQStackedWidget, self).__init__(parent)
        # Initialize Default Values
        self.fadeTransition = False
        self.slideTransition = False
        self.transitionDirection = Qt.Vertical
        self.transitionTime = 500
        self.fadeTime = 500
        self.transitionEasingCurve = "OutBack"
        self.fadeEasingCurve = "Linear"
        self.currentWidget = 0
        self.nextWidget = 0
        self._currentWidgetPosition = QPoint(0, 0)
        self.widgetActive = False

    ## Define properties for Qt Designer integration
    @Property(bool)
    def fadeTransition(self):
        return self._fadeTransition

    @fadeTransition.setter
    def fadeTransition(self, fadeState):
        if isinstance(fadeState, bool):
            self._fadeTransition = fadeState
        else:
            raise Exception("setFadeTransition() only accepts boolean variables")

    @Property(bool)
    def slideTransition(self):
        return self._slideTransition

    @slideTransition.setter
    def slideTransition(self, slideState):
        if isinstance(slideState, bool):
            self._slideTransition = slideState
        else:
            raise Exception("setSlideTransition() only accepts boolean variables")

    @Property(int)
    def transitionTime(self):
        return self._transitionTime

    @transitionTime.setter
    def transitionTime(self, time):
        self._transitionTime = time

    @Property(int)
    def fadeTime(self):
        return self._fadeTime

    @fadeTime.setter
    def fadeTime(self, time):
        self._fadeTime = time

    @Property(str)
    def transitionEasingCurve(self):
        return self._transitionEasingCurve

    @transitionEasingCurve.setter
    def transitionEasingCurve(self, curve):
        self._transitionEasingCurve = curve

    @Property(str)
    def fadeEasingCurve(self):
        return self._fadeEasingCurve

    @fadeEasingCurve.setter
    def fadeEasingCurve(self, curve):
        self._fadeEasingCurve = curve

    
    ## Function to update transition direction
    def setTransitionDirection(self, direction):
        self.transitionDirection = direction
    
    ## Function to update transition speed
    def setTransitionSpeed(self, speed):
        self.transitionTime = speed

    
    ## Function to update fade speed
    def setFadeSpeed(self, speed):
        self.fadeTime = speed

    
    ## Function to update transition easing curve
    def setTransitionEasingCurve(self, aesingCurve):
        self.transitionEasingCurve = aesingCurve

    
    ## Function to update fade easing curve
    def setFadeCurve(self, aesingCurve):
        self.fadeEasingCurve = aesingCurve

    
    ## Function to transition to previous widget
    @Slot()
    def slideToPreviousWidget(self):
        currentWidgetIndex = self.currentIndex()
        if currentWidgetIndex > 0:
            self.slideToWidgetIndex(currentWidgetIndex - 1)

    
    ## Function to transition to next widget
    @Slot()
    def slideToNextWidget(self):
        currentWidgetIndex = self.currentIndex()
        if currentWidgetIndex < (self.count() - 1):
            self.slideToWidgetIndex(currentWidgetIndex + 1)


    
    ## Function to transition to a given widget index
    def slideToWidgetIndex(self, index):
        if index > (self.count() - 1):
            index = index % self.count()
        elif index < 0:
            index = (index + self.count()) % self.count()
        if self.slideTransition:
            self.slideToWidget(self.widget(index))
        else:
            self.setCurrentIndex(index)

    
    ## Function to transition to a given widget
    def slideToWidget(self, newWidget):
        # If the widget is active, exit the function
        if self.widgetActive:
            return

        # Update widget active bool
        self.widgetActive = True

        # Get current and next widget index
        _currentWidgetIndex = self.currentIndex()
        _nextWidgetIndex = self.indexOf(newWidget)

        # If current widget index is equal to next widget index, exit function
        if _currentWidgetIndex == _nextWidgetIndex:
            self.widgetActive = False
            return

        anim_group = QParallelAnimationGroup(
            self, finished=self.animationDoneSlot
        )

        # Get X and Y position of QStackedWidget
        offsetX, offsetY = self.frameRect().width(), self.frameRect().height()
        # Set the next widget geometry
        self.widget(_nextWidgetIndex).setGeometry(self.frameRect())

        self.widget(_nextWidgetIndex).show()
        self.widget(_nextWidgetIndex).raise_()

        if self.slideTransition:
            # Animate transition
            # Set left right(horizontal) or up down(vertical) transition
            if not self.transitionDirection == Qt.Horizontal:
                if _currentWidgetIndex < _nextWidgetIndex:
                    # Down up transition
                    offsetX, offsetY = 0, -offsetY
                else:
                    # Up down transition
                    offsetX = 0
            else:
                # Right left transition
                if _currentWidgetIndex < _nextWidgetIndex:
                    offsetX, offsetY = -offsetX, 0
                else:
                    # Left right transition
                    offsetY = 0

            nextWidgetPosition = self.widget(_nextWidgetIndex).pos()
            currentWidgetPosition = self.widget(_currentWidgetIndex).pos()
            self._currentWidgetPosition = currentWidgetPosition

            offset = QPoint(offsetX, offsetY)
            self.widget(_nextWidgetIndex).move(nextWidgetPosition - offset)
            
            for index, start, end in zip(
                (_currentWidgetIndex, _nextWidgetIndex),
                (currentWidgetPosition, nextWidgetPosition - offset),
                (currentWidgetPosition + offset, nextWidgetPosition)
            ):
                animation = QPropertyAnimation(
                    self.widget(index),
                    b"pos",
                    duration=self.transitionTime,
                    easingCurve=returnAnimationEasingCurve(self.transitionEasingCurve),
                    startValue=start,
                    endValue=end,
                )
                anim_group.addAnimation(animation)

        # Play fade animation
        if self.fadeTransition:
            opacityEffect = QGraphicsOpacityEffect(self.widget(self.currentWidget))
            self.setGraphicsEffect(opacityEffect)
            opacityAni = QPropertyAnimation(opacityEffect, b'opacity', self.widget(self.currentWidget))
            opacityAni.setStartValue(0)
            opacityAni.setEndValue(1)
            opacityAni.setDuration(self.fadeTime)
            opacityAni.setEasingCurve(returnAnimationEasingCurve(self.fadeEasingCurve))
            opacityAni.finished.connect(opacityEffect.deleteLater)
            # opacityAni.start()

            anim_group.addAnimation(opacityAni)

        self.nextWidget = _nextWidgetIndex
        self.currentWidget = _currentWidgetIndex
        
        self.widgetActive = True
        # self.setCurrentIndex(self.nextWidget)
        anim_group.start(QAbstractAnimation.DeleteWhenStopped)

    
    ## Function to hide old widget and show new widget after animation is done
    @Slot()
    def animationDoneSlot(self):
        # self.widget(self.currentWidget).hide()
        self.setCurrentIndex(self.nextWidget)
        self.widget(self.currentWidget).move(self._currentWidgetPosition)
        self.widgetActive = False

    
    ## Function extending the QStackedWidget setCurrentWidget to animate transition
    @Slot()
    def setCurrentWidget(self, widget):
        currentIndex = self.currentIndex()
        nextIndex = self.indexOf(widget)
        if currentIndex == nextIndex and self.currentWidget == currentIndex:
            return
        
        # FadeWidgetTransition(self, self.widget(self.currentIndex()), self.widget(self.indexOf(widget)))
        self.slideToWidget(widget)
        # if not self.slideTransition:
        #     self.setCurrentIndex(0)
            
        if not self.slideTransition and not self.fadeTransition:
            self.setCurrentIndex(nextIndex)




"""
The primary objects for interfacing with iconify
"""

from typing import TYPE_CHECKING, MutableMapping, Optional, Tuple

from Custom_Widgets.iconify.anim import GlobalTick
from Custom_Widgets.iconify.path import findIcon
from Custom_Widgets.iconify.qt import QtCore, QtGui, QtSvg

if TYPE_CHECKING:
    from Custom_Widgets.iconify.anim import BaseAnimation
    from Custom_Widgets.iconify.qt import QtWidgets
    PixmapCacheKey = Tuple[Optional[str], QtCore.QSize, str, int, int]


class Icon(QtGui.QIcon):
    """
    The Iconify Icon which renders an svg image
    using the provided color & anim.
    """

    def __new__(
        cls,
        path,  # type: str
        color=None,  # type: Optional[QtGui.QColor]
        anim=None  # type: Optional[BaseAnimation]
    ):
        # type: (...) -> QtGui.QIcon
        """
        Return a QtGui.QIcon with extra convenience methods patched onto it.

        Patching a QIcon, rather than subclassing it, is required for the
        object to work correctly with Qt's model / view framework.

        Parameters
        ----------
        path : str
        color : Optional[QtGui.QColor]
        anim : Optional[BaseAnimation]

        Returns
        -------
        QtGui.QIcon
        """
        iconEngine = _IconEngine(path, color=color, anim=anim)
        icon = QtGui.QIcon(iconEngine)

        def _setAsButtonIcon(button):
            # type: (QtWidgets.QAbstractButton) -> None
            button.setIcon(icon)
            if iconEngine.animCount() > 1:
                GlobalTick.timeout.connect(button.update)
            else:
                if anim is not None:
                    anim.tick.connect(button.update)

        icon.setAsButtonIcon = _setAsButtonIcon
        icon.addState = iconEngine.addState
        icon.pixmapGenerator = iconEngine.pixmapGenerator
        icon.color = iconEngine.color
        icon.anim = iconEngine.anim
        icon.animCount = iconEngine.animCount
        return icon


class _IconEngine(QtGui.QIconEngine):
    """
    A QIconEngine which uses a PixmapGenerator for it's work.
    """

    def __init__(self, path, color=None, anim=None):
        # type: (str, Optional[QtGui.QColor], Optional[BaseAnimation]) -> None
        super(_IconEngine, self).__init__()
        self._defaultGenerator = PixmapGenerator(path, color=color, anim=anim)
        self._pixmapGenerators = {(QtGui.QIcon.Normal, QtGui.QIcon.Off):
                                  self._defaultGenerator}

    def __del__(self):
        # type: () -> None
        """
        Re-implemented to avoid this object being
        garbage collected unnecessarily.
        """

    def addState(
        self,
        path,  # type: str
        color=None,  # type: Optional[QtGui.QColor]
        anim=None,  # type: Optional[BaseAnimation]
        mode=QtGui.QIcon.Normal,  # type: QtGui.QIcon.Mode
        state=QtGui.QIcon.Off  # type: QtGui.QIcon.State
    ):
        # type: (...) -> None
        """
        Use the provided path, color and animation when an icon is requested
        for the identified mode and state.

        Parameters
        ----------
        path : str
        color : Optional[QtGui.QColor]
        anim : Optional[BaseAnimation]
        mode : QtGui.QIcon.Mode
        state : QtGui.QIcon.Stat
        """
        color = color or self.color()
        anim = anim or self.anim()
        generator = PixmapGenerator(path, color=color, anim=anim)
        self._pixmapGenerators[(mode, state)] = generator

    def pixmapGenerator(self, mode=QtGui.QIcon.Normal, state=QtGui.QIcon.Off):
        # type: (QtGui.QIcon.Mode, QtGui.QIcon.State) -> PixmapGenerator
        """
        Return the PixmapGenerator to use for the provided mode & state.

        If no PixmapGenerator has been defined for the provided mode & state
        combination, the PixmapGenerator created in the constructor is used.

        Parameters
        ----------
        mode : QtGui.QIcon.Mode
        state : QtGui.QIcon.State

        Returns
        -------
        PixmapGenerator
        """
        return self._pixmapGenerators.get(
            (mode, state),
            self._defaultGenerator,
        )

    def anim(
        self,
        mode=QtGui.QIcon.Normal,  # type: QtGui.QIcon.Mode
        state=QtGui.QIcon.Off,  # type: QtGui.QIcon.State
    ):
        # type: (...) -> Optional[BaseAnimation]
        """
        Return the animation used for the provided mode & state.

        Parameters
        ----------
        mode : QtGui.QIcon.Mode
        state : QtGui.QIcon.State

        Returns
        -------
        Optional[BaseAnimation]
        """
        return self.pixmapGenerator(mode=mode, state=state).anim()

    def animCount(self):
        # type: () -> int
        """
        Return the number of animations that are associated with the available
        states of this icon engine.

        Returns
        -------
        int
        """
        animCount = 0
        for pixmapGenerator in self._pixmapGenerators.values():
            if pixmapGenerator.anim() is not None:
                animCount += 1
        return animCount

    def color(
        self,
        mode=QtGui.QIcon.Normal,  # type: QtGui.QIcon.Mode
        state=QtGui.QIcon.Off,  # type: QtGui.QIcon.State
    ):
        # type: (...) -> Optional[QtGui.QColor]
        """
        Return the color used for the provided mode & state.

        Parameters
        ----------
        mode : QtGui.QIcon.Mode
        state : QtGui.QIcon.State

        Returns
        -------
        Optional[BaseAnimation]
        """
        return self.pixmapGenerator(mode=mode, state=state).color()

    def pixmap(
        self,
        size,  # type: QtCore.QSize
        mode,  # type: QtGui.QIcon.Mode
        state,  # type: QtGui.QIcon.State
    ):
        # type: (...) -> QtGui.QPixmap
        """
        Return a QPixmap to use for the identified mode & state, rendered at
        the provided size.

        Parameters
        ----------
        size : QtCore.QSize
        mode : QtGui.QIcon.Mode
        state : QtGui.QIcon.State

        Returns
        -------
        QtGui.QPixmap
        """
        return self.pixmapGenerator(mode=mode, state=state).pixmap(size)

    def paint(
        self,
        painter,  # type: QtCore.QPainter
        rect,  # type: QtCore.QRect
        mode,  # type: QtGui.QIcon.Mode
        state,  # type: QtGui.QIcon.State
    ):
        # type: (...) -> None
        painter.drawPixmap(
            rect.topLeft(), self.pixmap(rect.size(), mode, state)
        )


class PixmapGenerator(QtCore.QObject):
    """
    Responsible for rendering the svg image and applying the color and
    transform from the animation during the process.

    It's backed by a cache to ensure that redundant rendering does not happen.
    """

    _pixmapCache = {}  # type: MutableMapping[PixmapCacheKey, QtGui.QPixmap]

    def __init__(
        self,
        path,  # type: str
        color=None,  # type: Optional[QtGui.QColor]
        anim=None,  # type: Optional[BaseAnimation]
        parent=None,  # type: Optional[QtCore.QObject]
    ):
        # type: (...) -> None
        super(PixmapGenerator, self).__init__(parent=parent)
        self._path = findIcon(path)
        self._color = color  # type: Optional[QtGui.QColor]
        self._anim = anim  # type: Optional[BaseAnimation]

        self._renderer = QtSvg.QSvgRenderer(self._path)

    def path(self):
        # type: () -> str
        """
        Return the path to the image used by this PixmapGenerator.

        Returns
        -------
        str
        """
        return self._path

    def color(self):
        # type: () -> Optional[QtGui.QColor]
        """
        Return the color used by this PixmapGenerator.

        Returns
        -------
        Optional[QtGui.QColor]
        """
        return self._color

    def anim(self):
        # type: () -> Optional[BaseAnimation]
        """
        Return the animation used by this PixmapGenerator.

        Returns
        -------
        Optional[BaseAnimation]
        """
        return self._anim

    def pixmap(self, size):
        # type: (QtCore.QSize) -> QtGui.QPixmap
        """
        Render the svg file to a QPixmap, applying the color override and the
        animation transform if applicable.

        Parameters
        ----------
        size : QtCore.QSize

        Returns
        -------
        QtGui.QPixmap
        """
        color = self._color.rgb() if self._color else -1

        if self._anim is not None:
            key = (
                self._path, size, str(self._anim.__class__),
                self._anim.frame(), color
            )  # type: PixmapCacheKey
        else:
            key = (self._path, size, "", 0, color)

        if key in self._pixmapCache:
            return self._pixmapCache[key]

        image = QtGui.QImage(
            size,
            QtGui.QImage.Format_ARGB32_Premultiplied,
        )
        image.fill(QtCore.Qt.transparent)

        # Use the QSvgRenderer to draw the image
        painter = QtGui.QPainter(image)

        if self._anim:
            # Rotate the painter's co-ordinate space so
            # the image is correctly positioned.
            xfm = self._anim.transform(size)
            painter.setTransform(xfm)

        self._renderer.render(painter)
        painter.end()

        if self._color is not None:
            # Use the alpha channel on a solid colour image
            colorImage = QtGui.QImage(
                size,
                QtGui.QImage.Format_ARGB32_Premultiplied,
            )
            colorImage.fill(QtGui.QColor(self._color))
            try:
                colorImage.setAlphaChannel(image.alphaChannel())
            except AttributeError:
                if image.hasAlphaChannel():
                    colorImage.setAlphaChannel(image)

            image = colorImage

        pixmap = QtGui.QPixmap.fromImage(image)
        self._pixmapCache[key] = pixmap
        return pixmap
